'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os, zipfile
from distutils import errors
import bdist_portable_base


class bdist_winpenpack(bdist_portable_base.bdist_portable_base):
    
    description = 'create a winPenPack X-Software package'
    
    user_options = [
        ('bdist-base=', None, 
         'base directory for creating built distributions [build]'),
        ('dist-dir=', 'd', 'directory to put final package files in [dist]'),
        ('version=', None, 'version of the application'),
        ('license=', None, 'license (title) of the application'),
        ('url=', None, 'url of the application homepage'),
        ('filename=', None, 'filename of the application without extension'),
        ('date=', None, 'the release date')]
    
    def initialize_options(self):
        # pylint: disable=W0201
        self.bdist_base = 'build'
        self.dist_dir = 'dist'
        self.version = self.license = self.url = self.filename = self.date = None
        self.bdist_base_wpp = os.path.join(self.bdist_base, 
                                          'X-TaskCoach_%s_rev1'%self.version)
    
    def finalize_options(self):
        mandatoryOptions = [('version', 'the version of the application'),
                            ('license', 'the title of the license'),
                            ('url', 'the url of the application homepage'),
                            ('filename', 'the filename of the application without extension'),
                            ('date', 'the release date')]
        for option, description in mandatoryOptions:
            if not getattr(self, option):
                raise errors.DistutilsOptionError, \
                    'you must provide %s (--%s)'%(description, option)

    def run(self):
        self.create_winpenpack_paths()
        self.copy_launcher()
        self.copy_readme()
        self.copy_bin()
        self.zip()
        
    def create_winpenpack_paths(self):
        for pathComponents in [('Bin', 'TaskCoach'), 
                               ('Documents', 'TaskCoach'),
                               ('ReadMe',), 
                               ('User', 'TaskCoach')]:
            path = os.path.join(self.bdist_base_wpp, *pathComponents) # pylint: disable=W0142
            if not os.path.exists(path):
                os.makedirs(path)
                    
    def copy_launcher(self):
        launcher_src = os.path.join('build.in', 'winpenpack')
        self.copy_files(launcher_src, self.bdist_base_wpp)
               
    def copy_readme(self):
        readme_src = os.path.join('build.in', 'winpenpack', 'ReadMe')
        readme_dest = os.path.join(self.bdist_base_wpp, 'ReadMe')
        self.copy_files(readme_src, readme_dest)

    def copy_bin(self):
        srcdir = os.path.join(self.bdist_base, 'TaskCoach-%s-win32exe'%self.version)
        destdir = os.path.join(self.bdist_base_wpp, 'Bin', 'TaskCoach')
        self.copy_files(srcdir, destdir, copy_recursively=True)
        
    def zip(self):
        archive_filename = os.path.join(self.dist_dir, 'X-TaskCoach_%s_rev1.zip'%self.version)
        archive = zipfile.ZipFile(archive_filename, 'w', zipfile.ZIP_DEFLATED)
        for dirpath, _, filenames in os.walk(self.bdist_base_wpp):
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                arcname = filepath[len(self.bdist_base_wpp):]
                archive.write(filepath, arcname)
        archive.close()
        # For the buildbot
        print 'Generated %s' % archive_filename



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import config
from taskcoachlib.domain import task, date


class ViewFilterTestCase(test.TestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.list = task.TaskList()
        self.filter = task.filter.ViewFilter(self.list, treeMode=self.treeMode) # pylint: disable=E1101
        self.task = task.Task(subject='task')
        self.dueToday = task.Task(subject='due today', dueDateTime=date.Now().endOfDay())
        self.dueTomorrow = task.Task(subject='due tomorrow', 
            dueDateTime=date.Tomorrow().endOfDay())
        self.dueYesterday = task.Task(subject='due yesterday', 
            dueDateTime=date.Yesterday().endOfDay())
        self.child = task.Task(subject='child')
        
    def assertFilterShows(self, *tasks):
        self.assertEqual(len(tasks), len(self.filter))
        for eachTask in tasks:
            self.failUnless(eachTask in self.filter)
        
    def assertFilterIsEmpty(self):
        self.failIf(self.filter)


class ViewFilterTestsMixin(object):
    def testCreate(self):
        self.assertFilterIsEmpty()

    def testAddTask(self):
        self.filter.append(self.task)
        self.assertFilterShows(self.task)
        
    def testFilterCompletedTask(self):
        self.task.setCompletionDateTime()
        self.filter.append(self.task)
        self.assertFilterShows(self.task)
        self.filter.hideTaskStatus(task.status.completed)
        self.assertFilterIsEmpty()

    def testNrOfTasksPerStatusIsAffectedByFiltering(self):
        self.task.setCompletionDateTime()
        self.filter.append(self.task)
        self.filter.hideTaskStatus(task.status.completed)
        self.assertEqual(0, self.filter.nrOfTasksPerStatus()[task.status.completed])
                
    def testFilterCompletedTask_RootTasks(self):
        self.task.setCompletionDateTime()
        self.filter.append(self.task)
        self.filter.hideTaskStatus(task.status.completed)
        self.failIf(self.filter.rootItems())

    def testMarkTaskCompleted(self):
        self.filter.hideTaskStatus(task.status.completed)
        self.list.append(self.task)
        self.task.setCompletionDateTime()
        self.assertFilterIsEmpty()

    def testMarkTaskUncompleted(self):
        self.filter.hideTaskStatus(task.status.completed)
        self.task.setCompletionDateTime()
        self.list.append(self.task)
        self.task.setCompletionDateTime(date.DateTime())
        self.assertFilterShows(self.task)
        
    def testChangeCompletionDateOfAlreadyCompletedTask(self):
        self.filter.hideTaskStatus(task.status.completed)
        self.task.setCompletionDateTime()
        self.list.append(self.task)
        self.task.setCompletionDateTime(date.Tomorrow())
        self.assertFilterIsEmpty()
      
    def testFilterInactiveTask(self):
        self.task.setPlannedStartDateTime(date.Tomorrow())
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.inactive)
        self.assertFilterIsEmpty()
        
    def testFilterInactiveTask_ChangePlannedStartDateTime(self):
        self.task.setPlannedStartDateTime(date.Tomorrow())
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.inactive)
        self.task.setPlannedStartDateTime(date.Now() - date.ONE_SECOND)
        self.assertFilterShows(self.task)
        
    def testFilterInactiveTask_WhenPlannedStartDateTimePasses(self):
        plannedStart = date.Tomorrow()
        self.task.setPlannedStartDateTime(plannedStart)
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.inactive)
        oldNow = date.Now
        now = plannedStart + date.ONE_SECOND
        date.Now = lambda: now
        self.task.onTimeToStart()
        self.assertFilterShows(self.task)
        date.Now = oldNow

    def testMarkPrerequisiteCompletedWhileFilteringInactiveTasks(self):
        self.task.addPrerequisites([self.dueToday])
        self.dueToday.addDependencies([self.task])
        self.task.setPlannedStartDateTime(date.Now() - date.ONE_SECOND)
        self.dueToday.setPlannedStartDateTime(date.Now())
        self.filter.extend([self.dueToday, self.task])
        self.filter.hideTaskStatus(task.status.inactive)
        self.filter.hideTaskStatus(task.status.completed)
        self.assertFilterShows(self.dueToday)
        self.dueToday.setCompletionDateTime()
        self.assertFilterShows(self.task)
        
    def testAddPrerequisiteToActiveTaskWhileFilteringInactiveTasksShouldHideTask(self):
        for eachTask in (self.task, self.dueToday):
            eachTask.setPlannedStartDateTime(date.Now())
        self.filter.extend([self.dueToday, self.task])
        self.filter.hideTaskStatus(task.status.inactive)
        self.task.addPrerequisites([self.dueToday])
        self.assertFilterShows(self.dueToday)
        
    def testFilterLateTask(self):
        self.task.setPlannedStartDateTime(date.Yesterday())
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.late)
        self.assertFilterIsEmpty()

    def testFilterDueSoonTask(self):
        self.task.setDueDateTime(date.Now() + date.ONE_HOUR)
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.duesoon)
        self.assertFilterIsEmpty()
 
    def testFilterOverDueTask(self):
        self.task.setDueDateTime(date.Now() - date.ONE_HOUR)
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.overdue)
        self.assertFilterIsEmpty()
        
    def testFilterOverDueTaskWithActiveChild(self):
        self.child.setActualStartDateTime(date.Now())
        self.task.setDueDateTime(date.Now() - date.ONE_HOUR)
        self.task.addChild(self.child)
        self.list.append(self.task)
        self.filter.hideTaskStatus(task.status.overdue)
        if self.treeMode:
            self.assertFilterShows(self.task, self.child)
        else:
            self.assertFilterShows(self.child)


class ViewFilterInListModeTest(ViewFilterTestsMixin, ViewFilterTestCase):
    treeMode = False
            

class ViewFilterInTreeModeTest(ViewFilterTestsMixin, ViewFilterTestCase):
    treeMode = True
        
    def testFilterCompletedTasks(self):
        self.filter.hideTaskStatus(task.status.completed)
        child = task.Task()
        self.task.addChild(child)
        child.setParent(self.task)
        self.list.append(self.task)
        self.task.setCompletionDateTime()
        self.assertFilterIsEmpty()
        

class HideCompositeTasksTestCase(ViewFilterTestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.list = task.TaskList()
        self.filter = task.filter.ViewFilter(self.list, treeMode=self.treeMode) # pylint: disable=E1101
        self.task = task.Task(subject='task')
        self.child = task.Task(subject='child')
        self.task.addChild(self.child)
        self.filter.append(self.task)

    def _addTwoGrandChildren(self):
        # pylint: disable=W0201
        self.grandChild1 = task.Task(subject='grandchild 1')
        self.grandChild2 = task.Task(subject='grandchild 2')
        self.child.addChild(self.grandChild1)
        self.child.addChild(self.grandChild2)
        self.list.extend([self.grandChild1, self.grandChild2])


class HideCompositeTasksTestsMixin(object):
    def testTurnOn(self):
        self.filter.hideCompositeTasks()
        expectedTasks = (self.task, self.child) if self.filter.treeMode() else (self.child,)
        self.assertFilterShows(*expectedTasks) # pylint: disable=W0142

    def testTurnOff(self):
        self.filter.hideCompositeTasks()
        self.filter.hideCompositeTasks(False)
        self.assertFilterShows(self.task, self.child)
                
    def testAddChild(self):
        self.filter.hideCompositeTasks()
        grandChild = task.Task(subject='grandchild')
        self.list.append(grandChild)
        self.child.addChild(grandChild)
        expectedTasks = (self.task, self.child, grandChild) if self.filter.treeMode() else (grandChild,)
        self.assertFilterShows(*expectedTasks) # pylint: disable=W0142

    def testRemoveChild(self):
        self.filter.hideCompositeTasks()
        self.list.remove(self.child)
        self.assertFilterShows(self.task)

    def testAddTwoChildren(self):
        self.filter.hideCompositeTasks()
        self._addTwoGrandChildren()
        expectedTasks = (self.task, self.child, self.grandChild1, 
                         self.grandChild2) if self.filter.treeMode() else \
                        (self.grandChild1, self.grandChild2)
        self.assertFilterShows(*expectedTasks) # pylint: disable=W0142

    def testRemoveTwoChildren(self):
        self._addTwoGrandChildren()
        self.filter.hideCompositeTasks()
        self.list.removeItems([self.grandChild1, self.grandChild2])
        expectedTasks = (self.task, self.child) if self.filter.treeMode() else (self.child,)
        self.assertFilterShows(*expectedTasks) # pylint: disable=W0142


class HideCompositeTasksInListModeTest(HideCompositeTasksTestsMixin, 
                                       HideCompositeTasksTestCase):
    treeMode = False
            

class HideCompositeTasksInTreeModeTest(HideCompositeTasksTestsMixin, 
                                       HideCompositeTasksTestCase):
    treeMode = True



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.i18n import _
from taskcoachlib.domain import note
import base


class NewNoteCommand(base.NewItemCommand):
    singular_name = _('New note')
    
    def __init__(self, *args, **kwargs):
        subject = kwargs.pop('subject', _('New note'))
        description = kwargs.pop('description', '')
        attachments = kwargs.pop('attachments', [])
        categories = kwargs.get('categories',  None)
        super(NewNoteCommand, self).__init__(*args, **kwargs)
        self.items = self.notes = [note.Note(subject=subject,
            description=description, categories=categories, 
            attachments=attachments)]
        

class NewSubNoteCommand(base.NewSubItemCommand):
    plural_name = _('New subnotes')
    singular_name = _('New subnote of "%s"')

    def __init__(self, *args, **kwargs):
        subject = kwargs.pop('subject', _('New subnote'))
        description = kwargs.pop('description', '')
        attachments = kwargs.pop('attachments', [])
        categories = kwargs.get('categories',  None)
        super(NewSubNoteCommand, self).__init__(*args, **kwargs)
        self.items = self.notes = [parent.newChild(subject=subject,
            description=description, categories=categories,
            attachments=attachments) for parent in self.items]
        self.save_modification_datetimes()
                

class DeleteNoteCommand(base.DeleteCommand):
    plural_name = _('Delete notes')
    singular_name = _('Delete note "%s"')
    
    
class DragAndDropNoteCommand(base.OrderingDragAndDropCommand):
    plural_name = _('Drag and drop notes')
    singular_name = _('Drag and drop note "%s"')


class AddNoteCommand(base.BaseCommand):
    plural_name = _('Add note')
    singular_name = _('Add note to "%s"')

    def __init__(self, *args, **kwargs):
        self.owners = []
        super(AddNoteCommand, self).__init__(*args, **kwargs)
        self.owners = self.items
        self.items = self.__notes = [note.Note(subject=_('New note')) \
                                   for dummy in self.items]
        self.save_modification_datetimes()
        
    def modified_items(self):
        return self.owners

    def name_subject(self, newNote): # pylint: disable=W0613
        # Override to use the subject of the owner of the new note instead
        # of the subject of the new note itself, which wouldn't be very
        # interesting because it's something like 'New note'.
        return self.owners[0].subject()
    
    @patterns.eventSource
    def addNotes(self, event=None):
        for owner, note in zip(self.owners, self.__notes): # pylint: disable=W0621
            owner.addNote(note, event=event)

    @patterns.eventSource
    def removeNotes(self, event=None):
        for owner, note in zip(self.owners, self.__notes): # pylint: disable=W0621
            owner.removeNote(note, event=event)
    
    def do_command(self):
        super(AddNoteCommand, self).do_command()
        self.addNotes()
        
    def undo_command(self):
        super(AddNoteCommand, self).undo_command()
        self.removeNotes()
        
    def redo_command(self):
        super(AddNoteCommand, self).redo_command()
        self.addNotes()    


class AddSubNoteCommand(base.BaseCommand):
    plural_name = _('Add subnote')
    singular_name = _('Add subnote to "%s"')
    
    def __init__(self, *args, **kwargs):
        self.__owner = kwargs.pop('owner')
        self.__parents = []
        super(AddSubNoteCommand, self).__init__(*args, **kwargs)
        self.__parents = self.items
        self.__notes = kwargs.get('notes', [note.Note(subject=_('New subnote'),
                                                      parent=parent) \
                                            for parent in self.__parents])
        self.items = self.__notes
        self.save_modification_datetimes()
        
    def modified_items(self):
        return self.__parents + [self.__owner]
    
    @patterns.eventSource
    def addNotes(self, event=None):
        for parent, subnote in zip(self.__parents, self.__notes):
            parent.addChild(subnote, event=event)
            self.__owner.addNote(subnote, event=event)

    @patterns.eventSource
    def removeNotes(self, event=None):
        for parent, subnote in zip(self.__parents, self.__notes):
            parent.removeChild(subnote, event=event)
            self.__owner.removeNote(subnote, event=event)
    
    def do_command(self):
        super(AddSubNoteCommand, self).do_command()
        self.addNotes()
        
    def undo_command(self):
        super(AddSubNoteCommand, self).undo_command()
        self.removeNotes()
        
    def redo_command(self):
        super(AddSubNoteCommand, self).redo_command()
        self.addNotes()


class RemoveNoteCommand(base.BaseCommand):
    plural_name = _('Remove note')
    singular_name = _('Remove note from "%s"')
    
    def __init__(self, *args, **kwargs):
        self.__notes = kwargs.pop('notes')
        super(RemoveNoteCommand, self).__init__(*args, **kwargs)

    @patterns.eventSource
    def addNotes(self, event=None):
        kwargs = dict(event=event)
        for item in self.items:
            item.addNotes(*self.__notes, **kwargs) # pylint: disable=W0142
        
    @patterns.eventSource
    def removeNotes(self, event=None):
        # pylint: disable=W0142
        kwargs = dict(event=event)
        for item in self.items:
            for eachNote in self.__notes:
                if eachNote.parent():
                    eachNote.parent().removeChild(eachNote, **kwargs)
            item.removeNotes(*self.__notes, **kwargs) 
                
    def do_command(self):
        super(RemoveNoteCommand, self).do_command()
        self.removeNotes()
        
    def undo_command(self):
        super(RemoveNoteCommand, self).undo_command()
        self.addNotes()

    def redo_command(self):
        super(RemoveNoteCommand, self).redo_command()
        self.removeNotes()



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import meta, persistence, patterns, operating_system
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty import lockfile
from taskcoachlib.widgets import GetPassword
from taskcoachlib.workarounds import ExceptionAsUnicode
from taskcoachlib.gui.dialog import BackupManagerDialog
import wx
import os
import gc
import sys
import codecs
import traceback

try:
    from taskcoachlib.syncml import sync
except ImportError:  # pragma: no cover
    # Unsupported platform.
    pass


class IOController(object): 
    ''' IOController is responsible for opening, closing, loading,
    saving, and exporting files. It also presents the necessary dialogs
    to let the user specify what file to load/save/etc. '''

    def __init__(self, taskFile, messageCallback, settings, splash=None): 
        super(IOController, self).__init__()
        self.__taskFile = taskFile
        self.__messageCallback = messageCallback
        self.__settings = settings
        self.__splash = splash
        defaultPath = os.path.expanduser('~')
        self.__tskFileSaveDialogOpts = {'default_path': defaultPath, 
            'default_extension': 'tsk', 'wildcard': 
            _('%s files (*.tsk)|*.tsk|All files (*.*)|*') % meta.name}
        self.__tskFileOpenDialogOpts = {'default_path': defaultPath, 
            'default_extension': 'tsk', 'wildcard': 
            _('%s files (*.tsk)|*.tsk|Backup files (*.tsk.bak)|*.tsk.bak|'
              'All files (*.*)|*') % meta.name}
        self.__icsFileDialogOpts = {'default_path': defaultPath, 
            'default_extension': 'ics', 'wildcard': 
            _('iCalendar files (*.ics)|*.ics|All files (*.*)|*')}
        self.__htmlFileDialogOpts = {'default_path': defaultPath, 
            'default_extension': 'html', 'wildcard': 
            _('HTML files (*.html)|*.html|All files (*.*)|*')}
        self.__csvFileDialogOpts = {'default_path': defaultPath,
            'default_extension': 'csv', 'wildcard': 
            _('CSV files (*.csv)|*.csv|Text files (*.txt)|*.txt|'
              'All files (*.*)|*')}
        self.__todotxtFileDialogOpts = {'default_path': defaultPath,
            'default_extension': 'txt', 'wildcard':
            _('Todo.txt files (*.txt)|*.txt|All files (*.*)|*')}
        self.__errorMessageOptions = dict(caption=_('%s file error') % \
                                          meta.name, style=wx.ICON_ERROR)

    def syncMLConfig(self):
        return self.__taskFile.syncMLConfig()

    def setSyncMLConfig(self, config):
        self.__taskFile.setSyncMLConfig(config)

    def needSave(self):
        return self.__taskFile.needSave()

    def changedOnDisk(self):
        return self.__taskFile.changedOnDisk()

    def hasDeletedItems(self):
        return bool([task for task in self.__taskFile.tasks() if task.isDeleted()] + \
                    [note for note in self.__taskFile.notes() if note.isDeleted()])

    def purgeDeletedItems(self):
        self.__taskFile.tasks().removeItems([task for task in self.__taskFile.tasks() if task.isDeleted()])
        self.__taskFile.notes().removeItems([note for note in self.__taskFile.notes() if note.isDeleted()])

    def openAfterStart(self, commandLineArgs):
        ''' Open either the file specified on the command line, or the file
            the user was working on previously, or none at all. '''
        if commandLineArgs:
            filename = commandLineArgs[0].decode(sys.getfilesystemencoding())
        else:
            filename = self.__settings.get('file', 'lastfile')
        if filename:
            # Use CallAfter so that the main window is opened first and any 
            # error messages are shown on top of it
            wx.CallAfter(self.open, filename)
            
    def open(self, filename=None, showerror=wx.MessageBox, 
             fileExists=os.path.exists, breakLock=False, lock=True):
        if self.__taskFile.needSave():
            if not self.__saveUnsavedChanges():
                return
        if not filename:
            filename = self.__askUserForFile(_('Open'), 
                                             self.__tskFileOpenDialogOpts)
        if not filename:
            return
        self.__updateDefaultPath(filename)
        if fileExists(filename):
            self.__closeUnconditionally()
            self.__addRecentFile(filename)
            try:
                try:
                    self.__taskFile.load(filename, lock=lock, 
                                         breakLock=breakLock)
                except:
                    # Need to destroy splash screen first because it may 
                    # interfere with dialogs we show later on Mac OS X
                    if self.__splash:
                        self.__splash.Destroy()
                    raise
            except lockfile.LockTimeout:
                if breakLock:
                    if self.__askOpenUnlocked(filename): 
                        self.open(filename, showerror, lock=False)
                elif self.__askBreakLock(filename):
                    self.open(filename, showerror, breakLock=True)
                else:
                    return
            except lockfile.LockFailed:
                if self.__askOpenUnlocked(filename): 
                    self.open(filename, showerror, lock=False)
                else:
                    return
            except persistence.xml.reader.XMLReaderTooNewException:
                self.__showTooNewErrorMessage(filename, showerror)
                return
            except Exception:
                self.__showGenericErrorMessage(filename, showerror, showBackups=True)
                return
            self.__messageCallback(_('Loaded %(nrtasks)d tasks from '
                 '%(filename)s') % dict(nrtasks=len(self.__taskFile.tasks()), 
                                        filename=self.__taskFile.filename()))
        else:
            errorMessage = _("Cannot open %s because it doesn't exist") % filename
            # Use CallAfter on Mac OS X because otherwise the app will hang:
            if operating_system.isMac():
                wx.CallAfter(showerror, errorMessage, **self.__errorMessageOptions)
            else:
                showerror(errorMessage, **self.__errorMessageOptions)
            self.__removeRecentFile(filename)
            
    def merge(self, filename=None, showerror=wx.MessageBox):
        if not filename:
            filename = self.__askUserForFile(_('Merge'), 
                                             self.__tskFileOpenDialogOpts)
        if filename:
            try:
                self.__taskFile.merge(filename)
            except lockfile.LockTimeout:
                showerror(_('Cannot open %(filename)s\nbecause it is locked.') % \
                          dict(filename=filename),
                          **self.__errorMessageOptions)
                return
            except persistence.xml.reader.XMLReaderTooNewException:
                self.__showTooNewErrorMessage(filename, showerror)
                return
            except Exception:
                self.__showGenericErrorMessage(filename, showerror)
                return
            self.__messageCallback(_('Merged %(filename)s') % \
                                   dict(filename=filename))
            self.__addRecentFile(filename)

    def save(self, showerror=wx.MessageBox):
        if self.__taskFile.filename():
            if self._saveSave(self.__taskFile, showerror):
                return True
            else:
                return self.saveas(showerror=showerror)
        elif not self.__taskFile.isEmpty():
            return self.saveas(showerror=showerror)  # Ask for filename
        else:
            return False

    def mergeDiskChanges(self):
        self.__taskFile.mergeDiskChanges()

    def saveas(self, filename=None, showerror=wx.MessageBox, 
               fileExists=os.path.exists):
        if not filename:
            filename = self.__askUserForFile(_('Save as'), 
                self.__tskFileSaveDialogOpts,
                flag=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT, fileExists=fileExists)
            if not filename:
                return False  # User didn't enter a filename, cancel save
        if self._saveSave(self.__taskFile, showerror, filename):
            return True
        else:
            return self.saveas(showerror=showerror)  # Try again

    def saveselection(self, tasks, filename=None, showerror=wx.MessageBox,
                      TaskFileClass=persistence.TaskFile, 
                      fileExists=os.path.exists):
        if not filename:
            filename = self.__askUserForFile(_('Save selection'),
                self.__tskFileSaveDialogOpts, 
                flag=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT, fileExists=fileExists)
            if not filename:
                return False  # User didn't enter a filename, cancel save
        selectionFile = self._createSelectionFile(tasks, TaskFileClass)
        if self._saveSave(selectionFile, showerror, filename):
            return True
        else:
            return self.saveselection(tasks, showerror=showerror, 
                                      TaskFileClass=TaskFileClass)  # Try again
            
    def _createSelectionFile(self, tasks, TaskFileClass):
        selectionFile = TaskFileClass()
        # Add the selected tasks:
        selectionFile.tasks().extend(tasks)
        # Include categories used by the selected tasks:
        allCategories = set()
        for task in tasks:
            allCategories.update(task.categories())
        # Also include parents of used categories, recursively:
        for category in allCategories.copy():
            allCategories.update(category.ancestors())
        selectionFile.categories().extend(allCategories)
        return selectionFile
    
    def _saveSave(self, taskFile, showerror, filename=None):
        ''' Save the file and show an error message if saving fails. '''
        try:
            if filename:
                taskFile.saveas(filename)
            else:
                filename = taskFile.filename()
                taskFile.save()
            self.__showSaveMessage(taskFile)
            self.__addRecentFile(filename)
            return True
        except lockfile.LockTimeout:
            errorMessage = _('Cannot save %s\nIt is locked by another instance '
                             'of %s.\n') % (filename, meta.name)
            showerror(errorMessage, **self.__errorMessageOptions)
            return False
        except (OSError, IOError, lockfile.LockFailed), reason:
            errorMessage = _('Cannot save %s\n%s') % (filename, 
                           ExceptionAsUnicode(reason))
            showerror(errorMessage, **self.__errorMessageOptions)
            return False
        
    def saveastemplate(self, task):
        templates = persistence.TemplateList(self.__settings.pathToTemplatesDir())
        templates.addTemplate(task)
        templates.save()

    def importTemplate(self, showerror=wx.MessageBox):
        filename = self.__askUserForFile(_('Import template'),
            fileDialogOpts={'default_extension': 'tsktmpl', 
                            'wildcard': _('%s template files (*.tsktmpl)|'
                                          '*.tsktmpl') % meta.name})
        if filename:
            templates = persistence.TemplateList(self.__settings.pathToTemplatesDir())
            try:
                templates.copyTemplate(filename)
            except Exception, reason:  # pylint: disable=W0703
                errorMessage = _('Cannot import template %s\n%s') % (filename, 
                               ExceptionAsUnicode(reason))
                showerror(errorMessage, **self.__errorMessageOptions)
            
    def close(self, force=False):
        if self.__taskFile.needSave():
            if force:
                # No user interaction, since we're forced to close right now.
                if self.__taskFile.filename():
                    self._saveSave(self.__taskFile, 
                                   lambda *args, **kwargs: None)
                else:
                    pass  # No filename, we cannot ask, give up...
            else:
                if not self.__saveUnsavedChanges():
                    return False
        self.__closeUnconditionally()
        return True
    
    def export(self, title, fileDialogOpts, writerClass, viewer, selectionOnly, 
               openfile=codecs.open, showerror=wx.MessageBox, filename=None, 
               fileExists=os.path.exists, **kwargs):
        filename = filename or self.__askUserForFile(title, fileDialogOpts, 
            flag=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT, fileExists=fileExists)
        if filename:
            fd = self.__openFileForWriting(filename, openfile, showerror)
            if fd is None:
                return False
            count = writerClass(fd, filename).write(viewer, self.__settings, 
                                                    selectionOnly, **kwargs)
            fd.close()
            self.__messageCallback(_('Exported %(count)d items to '
                                     '%(filename)s') % dict(count=count, 
                                                            filename=filename))
            return True
        else:
            return False

    def exportAsHTML(self, viewer, selectionOnly=False, separateCSS=False,
                     columns=None, openfile=codecs.open, 
                     showerror=wx.MessageBox, filename=None, 
                     fileExists=os.path.exists):
        return self.export(_('Export as HTML'), self.__htmlFileDialogOpts, 
            persistence.HTMLWriter, viewer, selectionOnly, openfile, showerror, 
            filename, fileExists, separateCSS=separateCSS, columns=columns)

    def exportAsCSV(self, viewer, selectionOnly=False, 
                    separateDateAndTimeColumns=False, columns=None,
                    fileExists=os.path.exists):
        return self.export(_('Export as CSV'), self.__csvFileDialogOpts, 
            persistence.CSVWriter, viewer, selectionOnly, 
            separateDateAndTimeColumns=separateDateAndTimeColumns, 
            columns=columns, fileExists=fileExists)
        
    def exportAsICalendar(self, viewer, selectionOnly=False, 
                          fileExists=os.path.exists):
        return self.export(_('Export as iCalendar'),
            self.__icsFileDialogOpts, persistence.iCalendarWriter, viewer, 
            selectionOnly, fileExists=fileExists)
        
    def exportAsTodoTxt(self, viewer, selectionOnly=False,
                        fileExists=os.path.exists):
        return self.export(_('Export as Todo.txt'),
            self.__todotxtFileDialogOpts, persistence.TodoTxtWriter, viewer,
            selectionOnly, fileExists=fileExists)

    def importCSV(self, **kwargs):
        persistence.CSVReader(self.__taskFile.tasks(),
                              self.__taskFile.categories()).read(**kwargs)
                              
    def importTodoTxt(self, filename):
        persistence.TodoTxtReader(self.__taskFile.tasks(),
                                  self.__taskFile.categories()).read(filename)

    def synchronize(self):
        doReset = False
        while True:
            password = GetPassword('Task Coach', 'SyncML', reset=doReset)
            if not password:
                break

            synchronizer = sync.Synchronizer(self.__syncReport,
                                         self.__taskFile, password)
            try:
                synchronizer.synchronize()
            except sync.AuthenticationFailure:
                doReset = True
            else:
                self.__messageCallback(_('Finished synchronization'))
                break
            finally:
                synchronizer.Destroy()

    def filename(self):
        return self.__taskFile.filename()

    def __syncReport(self, msg):
        wx.MessageBox(msg, _('Synchronization status'), 
                      style=wx.OK | wx.ICON_ERROR)

    def __openFileForWriting(self, filename, openfile, showerror, mode='w', 
                             encoding='utf-8'):
        try:
            return openfile(filename, mode, encoding)
        except IOError, reason:
            errorMessage = _('Cannot open %s\n%s') % (filename, 
                           ExceptionAsUnicode(reason))
            showerror(errorMessage, **self.__errorMessageOptions)
            return None
        
    def __addRecentFile(self, fileName):
        recentFiles = self.__settings.getlist('file', 'recentfiles')
        if fileName in recentFiles:
            recentFiles.remove(fileName)
        recentFiles.insert(0, fileName)
        maximumNumberOfRecentFiles = self.__settings.getint('file', 
                                                            'maxrecentfiles')
        recentFiles = recentFiles[:maximumNumberOfRecentFiles]
        self.__settings.setlist('file', 'recentfiles', recentFiles)
        
    def __removeRecentFile(self, fileName):
        recentFiles = self.__settings.getlist('file', 'recentfiles')
        if fileName in recentFiles:
            recentFiles.remove(fileName)
            self.__settings.setlist('file', 'recentfiles', recentFiles)
        
    def __askUserForFile(self, title, fileDialogOpts, flag=wx.FD_OPEN, 
                         fileExists=os.path.exists):
        filename = wx.FileSelector(title, flags=flag, 
                                   **fileDialogOpts)  # pylint: disable=W0142
        if filename and (flag & wx.FD_SAVE):
            # On Ubuntu, the default extension is not added automatically to
            # a filename typed by the user. Add the extension if necessary.
            extension = os.path.extsep + fileDialogOpts['default_extension']
            if not filename.endswith(extension):
                filename += extension
                if fileExists(filename):
                    return self.__askUserForOverwriteConfirmation(filename, title, 
                                                                  fileDialogOpts)
        return filename
    
    def __askUserForOverwriteConfirmation(self, filename, title, 
                                          fileDialogOpts):
        result = wx.MessageBox(_('A file named %s already exists.\n'
                                 'Do you want to replace it?') % filename,
            title, 
            style=wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION | wx.NO_DEFAULT)
        if result == wx.YES:
            extensions = { 'Todo.txt': '.txt' }
            for auto in set(self.__settings.getlist('file', 'autoimport') + self.__settings.getlist('file', 'autoexport')):
                autoName = os.path.splitext(filename)[0] + extensions[auto]
                if os.path.exists(autoName):
                    os.remove(autoName)
                if os.path.exists(autoName + '-meta'):
                    os.remove(autoName + '-meta')
            return filename
        elif result == wx.NO:
            return self.__askUserForFile(title, fileDialogOpts, 
                                         flag=wx.FD_SAVE | \
                                              wx.FD_OVERWRITE_PROMPT)
        else:
            return None

    def __saveUnsavedChanges(self):
        result = wx.MessageBox(_('You have unsaved changes.\n'
            'Save before closing?'), _('%s: save changes?') % meta.name,
            style=wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION | wx.YES_DEFAULT)
        if result == wx.YES:
            if not self.save():
                return False
        elif result == wx.CANCEL:
            return False
        return True
    
    def __askBreakLock(self, filename):
        result = wx.MessageBox(_('''Cannot open %s because it is locked.

This means either that another instance of TaskCoach
is running and has this file opened, or that a previous
instance of Task Coach crashed. If no other instance is
running, you can safely break the lock.

Break the lock?''') % filename,
            _('%s: file locked') % meta.name,
            style=wx.YES_NO | wx.ICON_QUESTION | wx.NO_DEFAULT)
        return result == wx.YES
    
    def __askOpenUnlocked(self, filename):
        result = wx.MessageBox(_('Cannot acquire a lock because locking is not '
            'supported\non the location of %s.\n'
            'Open %s unlocked?') % (filename, filename), 
             _('%s: file locked') % meta.name,
            style=wx.YES_NO | wx.ICON_QUESTION | wx.NO_DEFAULT)
        return result == wx.YES
    
    def __closeUnconditionally(self):
        self.__messageCallback(_('Closed %s') % self.__taskFile.filename())
        self.__taskFile.close()
        patterns.CommandHistory().clear()
        gc.collect()

    def __showSaveMessage(self, savedFile):    
        self.__messageCallback(_('Saved %(nrtasks)d tasks to %(filename)s') % \
            {'nrtasks': len(savedFile.tasks()), 
             'filename': savedFile.filename()})
        
    def __showTooNewErrorMessage(self, filename, showerror):
        showerror(_('Cannot open %(filename)s\n'
                    'because it was created by a newer version of %(name)s.\n'
                    'Please upgrade %(name)s.') % \
            dict(filename=filename, name=meta.name),
            **self.__errorMessageOptions)
        
    def __showGenericErrorMessage(self, filename, showerror, showBackups=False):
        sys.stderr.write(''.join(traceback.format_exception(*sys.exc_info())))
        limitedException = ''.join(traceback.format_exception(*sys.exc_info(), 
                                                              limit=10))
        message = _('Error while reading %s:\n') % filename + \
                    limitedException
        man = persistence.BackupManifest(self.__settings)
        if showBackups and man.hasBackups(filename):
            message += u'\n' + _('The backup manager will now open to allow you to restore\nan older version of this file.')
        showerror(message, **self.__errorMessageOptions)

        if showBackups and man.hasBackups(filename):
            dlg = BackupManagerDialog(None, self.__settings, filename)
            try:
                if dlg.ShowModal() == wx.ID_OK:
                    wx.CallAfter(self.open, dlg.restoredFilename())
            finally:
                dlg.Destroy()

    def __updateDefaultPath(self, filename):
        for options in [self.__tskFileOpenDialogOpts, 
                        self.__tskFileSaveDialogOpts,
                        self.__csvFileDialogOpts,
                        self.__icsFileDialogOpts, 
                        self.__htmlFileDialogOpts]:
            options['default_path'] = os.path.dirname(filename)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import optparse
from taskcoachlib import meta


class OptionParser(optparse.OptionParser, object):
    def __init__(self, *args, **kwargs):
        super(OptionParser, self).__init__(*args, **kwargs)
        self.__addOptionGroups()
        self.__addOptions()

    def __addOptionGroups(self):
        self.__getAndAddOptions('OptionGroup', self.add_option_group)
        
    def __addOptions(self):
        self.__getAndAddOptions('Option', self.add_option)
        
    def __getAndAddOptions(self, suffix, addOption):
        for getOption in self.__methodsEndingWith(suffix):
            addOption(getOption(self))

    def __methodsEndingWith(self, suffix):
        return [method for name, method in vars(self.__class__).items() if
                name.endswith(suffix)]

                
class OptionGroup(optparse.OptionGroup, object):
    pass
    
    
class ApplicationOptionParser(OptionParser):
    def __init__(self, *args, **kwargs):
        kwargs['usage'] = 'usage: %prog [options] [.tsk file]'
        kwargs['version'] = '%s %s'%(meta.data.name, meta.data.version)
        super(ApplicationOptionParser, self).__init__(*args, **kwargs)
        
    def profileOption(self):
        return optparse.Option('--profile', default=False, 
            action='store_true', help=optparse.SUPPRESS_HELP)
 
    def profile_skipstartOption(self):
        return optparse.Option('-s', '--skipstart', default=False, 
            action='store_true', help=optparse.SUPPRESS_HELP)

    def iniOption(self):
        return optparse.Option('-i', '--ini', dest='inifile',
            help='use the specified INIFILE for storing settings')
        
    def languageOption(self):
        return optparse.Option('-l', '--language', dest='language', 
            type='choice', choices=sorted([lang for (lang, enabled) in \
                meta.data.languages.values() if lang is not None] + ['en']),
            help='use the specified LANGUAGE for the GUI (e.g. "nl" or "fr"')

    def poOption(self):
        return optparse.Option('-p', '--po', dest='pofile',
            help='use the specified POFILE for translation of the GUI') 



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>
Copyright (C) 2008 Thomas Sonne Olesen <tpo@sonnet.dk>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import math
import wx.lib.agw.piectrl
from taskcoachlib import command, widgets, domain, render
from taskcoachlib.domain import task, date
from taskcoachlib.gui import uicommand, menu, dialog
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib.thirdparty.wxScheduler import wxSCHEDULER_TODAY, wxFancyDrawer
from taskcoachlib.thirdparty import smartdatetimectrl as sdtc
from taskcoachlib.widgets import CalendarConfigDialog, HierarchicalCalendarConfigDialog
import base
import inplace_editor
import mixin
import refresher 


class DueDateTimeCtrl(inplace_editor.DateTimeCtrl):
    def __init__(self, parent, wxId, item, column, owner, value, **kwargs):
        kwargs['relative'] = True
        kwargs['startDateTime'] = item.GetData().plannedStartDateTime()
        super(DueDateTimeCtrl, self).__init__(parent, wxId, item, column, owner, value, **kwargs)
        sdtc.EVT_TIME_CHOICES_CHANGE(self._dateTimeCtrl, self.OnChoicesChange)
        self._dateTimeCtrl.LoadChoices(item.GetData().settings.get('feature', 'sdtcspans'))

    def OnChoicesChange(self, event):
        self.item().GetData().settings.settext('feature', 'sdtcspans', event.GetValue())


class TaskViewerStatusMessages(object):
    template1 = _('Tasks: %d selected, %d visible, %d total')
    template2 = _('Status: %d overdue, %d late, %d inactive, %d completed')
    
    def __init__(self, viewer):
        super(TaskViewerStatusMessages, self).__init__()
        self.__viewer = viewer
        self.__presentation = viewer.presentation()
    
    def __call__(self):
        count = self.__presentation.observable(recursive=True).nrOfTasksPerStatus()
        return self.template1 % (len(self.__viewer.curselection()), 
                                 self.__viewer.nrOfVisibleTasks(), 
                                 self.__presentation.originalLength()), \
               self.template2 % (count[task.status.overdue], 
                                 count[task.status.late],
                                 count[task.status.inactive], 
                                 count[task.status.completed])


class BaseTaskViewer(mixin.SearchableViewerMixin,  # pylint: disable=W0223
                     mixin.FilterableViewerForTasksMixin,
                     base.CategorizableViewerMixin,
                     base.WithAttachmentsViewerMixin,
                     base.TreeViewer):        
    def __init__(self, *args, **kwargs):
        super(BaseTaskViewer, self).__init__(*args, **kwargs)
        self.statusMessages = TaskViewerStatusMessages(self)
        self.__registerForAppearanceChanges()
        wx.CallAfter(self.__DisplayBalloon)

    def __DisplayBalloon(self):
        if self.toolbar.getToolIdByCommand('ViewerHideTasks_completed') != wx.ID_ANY and self.toolbar.IsShownOnScreen() and \
            hasattr(wx.GetTopLevelParent(self), 'AddBalloonTip'):
            wx.GetTopLevelParent(self).AddBalloonTip(self.settings, 'filtershiftclick', self.toolbar,
                        getRect=lambda: self.toolbar.GetToolRect(self.toolbar.getToolIdByCommand('ViewerHideTasks_completed')),
                        message=_('''Shift-click on a filter tool to see only tasks belonging to the corresponding status'''))

    def __registerForAppearanceChanges(self):
        for appearance in ('font', 'fgcolor', 'bgcolor', 'icon'):
            appearanceSettings = ['settings.%s.%s' % (appearance, setting) for setting in 'activetasks',\
                                  'inactivetasks', 'completedtasks', 'duesoontasks', 'overduetasks', 'latetasks'] 
            for appearanceSetting in appearanceSettings:
                pub.subscribe(self.onAppearanceSettingChange, appearanceSetting)
        self.registerObserver(self.onAttributeChanged_Deprecated,
                              eventType=task.Task.appearanceChangedEventType())
        pub.subscribe(self.onAttributeChanged, task.Task.prerequisitesChangedEventType())
        pub.subscribe(self.refresh, 'powermgt.on')

    def detach(self):
        super(BaseTaskViewer, self).detach()
        self.statusMessages = None # Break cycle

    def _renderTimeSpent(self, *args, **kwargs):
        if self.settings.getboolean('feature', 'decimaltime'):
            return render.timeSpentDecimal(*args, **kwargs)
        return render.timeSpent(*args, **kwargs)

    def onAppearanceSettingChange(self, value):  # pylint: disable=W0613
        if self:
            wx.CallAfter(self.refresh)  # Let domain objects update appearance first
        # Show/hide status in toolbar may change too
        self.toolbar.loadPerspective(self.toolbar.perspective(), cache=False)

    def domainObjectsToView(self):
        return self.taskFile.tasks()

    def isShowingTasks(self): 
        return True

    def createFilter(self, taskList):
        tasks = domain.base.DeletedFilter(taskList)
        return super(BaseTaskViewer, self).createFilter(tasks)

    def nrOfVisibleTasks(self):
        # Make this overridable for viewers where the widget does not show all
        # items in the presentation, i.e. the widget does filtering on its own.
        return len(self.presentation())


class BaseTaskTreeViewer(BaseTaskViewer):  # pylint: disable=W0223
    defaultTitle = _('Tasks')
    defaultBitmap = 'led_blue_icon'
    
    def __init__(self, *args, **kwargs):
        super(BaseTaskTreeViewer, self).__init__(*args, **kwargs)
        if kwargs.get('doRefresh', True):
            self.secondRefresher = refresher.SecondRefresher(self,
                                                             task.Task.trackingChangedEventType())
            self.minuteRefresher = refresher.MinuteRefresher(self)
        else:
            self.secondRefresher = self.minuteRefresher = None
        
    def detach(self):
        super(BaseTaskTreeViewer, self).detach()
        if hasattr(self, "secondRefresher") and self.secondRefresher:
            self.secondRefresher.stopClock()
            self.secondRefresher.removeInstance()
            del self.secondRefresher
        if hasattr(self, "minuteRefresher") and self.minuteRefresher:
            self.minuteRefresher.stopClock()
            del self.minuteRefresher
        
    def newItemDialog(self, *args, **kwargs):
        kwargs['categories'] = self.taskFile.categories().filteredCategories()
        return super(BaseTaskTreeViewer, self).newItemDialog(*args, **kwargs)
    
    def editItemDialog(self, items, bitmap, columnName='', 
                       items_are_new=False):
        if isinstance(items[0], task.Task):
            return super(BaseTaskTreeViewer, self).editItemDialog(items, bitmap, 
                columnName=columnName, items_are_new=items_are_new)
        else:
            return dialog.editor.EffortEditor(wx.GetTopLevelParent(self),
                items, self.settings, self.taskFile.efforts(), self.taskFile,  
                bitmap=bitmap, items_are_new=items_are_new)
            
    def itemEditorClass(self):
        return dialog.editor.TaskEditor
    
    def newItemCommandClass(self):
        return command.NewTaskCommand
    
    def newSubItemCommandClass(self):
        return command.NewSubTaskCommand
    
    def newSubItemCommand(self):
        kwargs = dict() 
        if self.__shouldPresetPlannedStartDateTime():
            kwargs['plannedStartDateTime'] = task.Task.suggestedPlannedStartDateTime()
        if self.__shouldPresetDueDateTime():
            kwargs['dueDateTime'] = task.Task.suggestedDueDateTime()
        if self.__shouldPresetActualStartDateTime():
            kwargs['actualStartDateTime'] = task.Task.suggestedActualStartDateTime()
        if self.__shouldPresetCompletionDateTime():
            kwargs['completionDateTime'] = task.Task.suggestedCompletionDateTime()
        if self.__shouldPresetReminderDateTime():
            kwargs['reminder'] = task.Task.suggestedReminderDateTime()
        # pylint: disable=W0142
        return self.newSubItemCommandClass()(self.presentation(), 
                                             self.curselection(), **kwargs)

    def __shouldPresetPlannedStartDateTime(self):
        return self.settings.get('view', 'defaultplannedstartdatetime').startswith('preset')
            
    def __shouldPresetDueDateTime(self):
        return self.settings.get('view', 'defaultduedatetime').startswith('preset')

    def __shouldPresetActualStartDateTime(self):
        return self.settings.get('view', 'defaultactualstartdatetime').startswith('preset')
    
    def __shouldPresetCompletionDateTime(self):
        return self.settings.get('view', 'defaultcompletiondatetime').startswith('preset')
            
    def __shouldPresetReminderDateTime(self):
        return self.settings.get('view', 'defaultreminderdatetime').startswith('preset')
           
    def deleteItemCommand(self):
        return command.DeleteTaskCommand(self.presentation(), self.curselection(),
                  shadow=self.settings.getboolean('feature', 'syncml'))    

    def createTaskPopupMenu(self):
        return menu.TaskPopupMenu(self.parent, self.settings,
                                  self.presentation(), self.taskFile.efforts(),
                                  self.taskFile.categories(), self)

    def createCreationToolBarUICommands(self):
        return (uicommand.TaskNew(taskList=self.presentation(),
                                  settings=self.settings),
                uicommand.NewSubItem(viewer=self),
                uicommand.TaskNewFromTemplateButton(taskList=self.presentation(),
                                                    settings=self.settings,
                                                    bitmap='newtmpl')) + \
            super(BaseTaskTreeViewer, self).createCreationToolBarUICommands()
    
    def createActionToolBarUICommands(self):
        uiCommands = (uicommand.AddNote(settings=self.settings, viewer=self),
                      uicommand.TaskMarkInactive(settings=self.settings, viewer=self),
                      uicommand.TaskMarkActive(settings=self.settings, viewer=self),
                      uicommand.TaskMarkCompleted(settings=self.settings, viewer=self))
        uiCommands += (
            # EffortStart needs a reference to the original (task) list to
            # be able to stop tracking effort for tasks that are already 
            # being tracked, but that might be filtered in the viewer's 
            # presentation.
            None,
            uicommand.EffortStart(viewer=self, 
                                  taskList=self.taskFile.tasks()),
            uicommand.EffortStop(viewer=self,
                                 effortList=self.taskFile.efforts(),
                                 taskList=self.taskFile.tasks()))
        return uiCommands + super(BaseTaskTreeViewer, self).createActionToolBarUICommands()
    
    def createModeToolBarUICommands(self):
        hideUICommands = tuple([uicommand.ViewerHideTasks(taskStatus=status,
                                                          settings=self.settings,
                                                          viewer=self) \
                                for status in task.Task.possibleStatuses()])
        otherModeUICommands = super(BaseTaskTreeViewer, self).createModeToolBarUICommands()
        separator = (None,) if otherModeUICommands else ()
        return hideUICommands + separator + otherModeUICommands

    def iconName(self, item, isSelected):
        return item.selectedIcon(recursive=True) if isSelected else item.icon(recursive=True)

    def getItemTooltipData(self, task):  # pylint: disable=W0621
        result = [(self.iconName(task, task in self.curselection()), 
                   [self.getItemText(task)])]
        if task.notes():
            result.append(('note_icon', sorted([note.subject() for note in task.notes()])))
        if task.attachments():
            result.append(('paperclip_icon',
                sorted([unicode(attachment) for attachment in task.attachments()])))
        return result + super(BaseTaskTreeViewer, self).getItemTooltipData(task)

    def label(self, task):  # pylint: disable=W0621
        return self.getItemText(task)


class RootNode(object):
    def __init__(self, tasks):
        self.tasks = tasks
        
    def subject(self):
        return ''
    
    def children(self, recursive=False):
        if recursive:
            return self.tasks[:]
        else:
            return self.tasks.rootItems()

    # pylint: disable=W0613
    
    def foregroundColor(self, *args, **kwargs): 
        return None

    def backgroundColor(self, *args, **kwargs):
        return None
    
    def font(self, *args, **kwargs):
        return None

    def completed(self, *args, **kwargs):
        return False

    late = dueSoon = inactive = overdue = isBeingTracked = completed


class SquareMapRootNode(RootNode):
    def __getattr__(self, attr):
        def getTaskAttribute(recursive=True):
            if recursive:
                return max(sum((getattr(task, attr)(recursive=True) for task in self.children()),
                               self.__zero), 
                           self.__zero)
            else:
                return self.__zero

        self.__zero = date.TimeDelta() if attr in ('budget', 'budgetLeft', 'timeSpent') else 0  # pylint: disable=W0201
        return getTaskAttribute


class TimelineRootNode(RootNode):
    def children(self, recursive=False):
        children = super(TimelineRootNode, self).children(recursive)
        children.sort(key=lambda task: task.plannedStartDateTime())
        return children
    
    def parallel_children(self, recursive=False):
        return self.children(recursive)

    def sequential_children(self):
        return []

    def plannedStartDateTime(self, recursive=False):  # pylint: disable=W0613
        plannedStartDateTimes = [item.plannedStartDateTime(recursive=True) for item in self.parallel_children()]
        plannedStartDateTimes = [dt for dt in plannedStartDateTimes if dt != date.DateTime()]
        if not plannedStartDateTimes:
            plannedStartDateTimes.append(date.Now())
        return min(plannedStartDateTimes)
    
    def dueDateTime(self, recursive=False):  # pylint: disable=W0613
        dueDateTimes = [item.dueDateTime(recursive=True) for item in self.parallel_children()]
        dueDateTimes = [dt for dt in dueDateTimes if dt != date.DateTime()]
        if not dueDateTimes:
            dueDateTimes.append(date.Tomorrow())    
        return max(dueDateTimes)
    

class TimelineViewer(BaseTaskTreeViewer):
    defaultTitle = _('Timeline')
    defaultBitmap = 'timelineviewer'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'timelineviewer')
        super(TimelineViewer, self).__init__(*args, **kwargs)
        for eventType in (task.Task.subjectChangedEventType(),
                          task.Task.plannedStartDateTimeChangedEventType(),
                          task.Task.dueDateTimeChangedEventType(),
                          task.Task.completionDateTimeChangedEventType()):
            if eventType.startswith('pubsub'):
                pub.subscribe(self.onAttributeChanged, eventType)
            else:
                self.registerObserver(self.onAttributeChanged_Deprecated, 
                                      eventType)
        
    def createWidget(self):
        self.rootNode = TimelineRootNode(self.presentation())  # pylint: disable=W0201
        itemPopupMenu = self.createTaskPopupMenu()
        self._popupMenus.append(itemPopupMenu)
        return widgets.Timeline(self, self.rootNode, self.onSelect, self.onEdit,
                                itemPopupMenu)

    def onEdit(self, item):
        edit = uicommand.Edit(viewer=self)
        edit(item)
        
    def curselection(self):
        # Override curselection, because there is no need to translate indices
        # back to domain objects. Our widget already returns the selected domain
        # object itself.
        return self.widget.curselection()
    
    def bounds(self, item):
        times = [self.start(item), self.stop(item)]
        for child in self.parallel_children(item) + self.sequential_children(item):
            times.extend(self.bounds(child))
        times = [time for time in times if time is not None]
        return (min(times), max(times)) if times else []
 
    def start(self, item, recursive=False):
        try:
            start = item.plannedStartDateTime(recursive=recursive)
            if start == date.DateTime():
                return None
        except AttributeError:
            start = item.getStart()
        return start.toordinal()

    def stop(self, item, recursive=False):
        try:
            if item.completed():
                stop = item.completionDateTime(recursive=recursive)
            else:
                stop = item.dueDateTime(recursive=recursive)
            if stop == date.DateTime():
                return None   
            else:
                stop += date.ONE_DAY
        except AttributeError:
            stop = item.getStop()
            if not stop:
                return None
        return stop.toordinal() 

    def sequential_children(self, item):
        try:
            return item.efforts()
        except AttributeError:
            return []

    def parallel_children(self, item, recursive=False):
        try:
            children = [child for child in item.children(recursive=recursive) \
                        if child in self.presentation()]
            children.sort(key=lambda task: task.plannedStartDateTime())
            return children
        except AttributeError:
            return []

    def foreground_color(self, item, depth=0):  # pylint: disable=W0613
        return item.foregroundColor(recursive=True)
          
    def background_color(self, item, depth=0):  # pylint: disable=W0613
        return item.backgroundColor(recursive=True)
    
    def font(self, item, depth=0):  # pylint: disable=W0613
        return item.font(recursive=True)

    def icon(self, item, isSelected=False):
        bitmap = self.iconName(item, isSelected)
        return wx.ArtProvider_GetIcon(bitmap, wx.ART_MENU, (16, 16))
    
    def now(self):
        return date.Now().toordinal()
    
    def nowlabel(self):
        return _('Now')

    def getItemTooltipData(self, item):
        if isinstance(item, task.Task):
            result = super(TimelineViewer, self).getItemTooltipData(item)
        else:
            result = [(None, [render.dateTimePeriod(item.getStart(), item.getStop(), humanReadable=True)])]
            if item.description(): 
                result.append((None, [line.rstrip('\n') for line in item.description().split('\n')]))     
        return result


class SquareTaskViewer(BaseTaskTreeViewer):
    defaultTitle = _('Task square map')
    defaultBitmap = 'squaremapviewer'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'squaretaskviewer')
        self.__orderBy = 'revenue'
        self.__transformTaskAttribute = lambda x: x
        self.__zero = 0
        self.renderer = dict(budget=render.budget, timeSpent=self._renderTimeSpent, 
                        fixedFee=render.monetaryAmount, 
                        revenue=render.monetaryAmount,
                        priority=render.priority)
        super(SquareTaskViewer, self).__init__(*args, **kwargs)
        sortKeys = eval(self.settings.get(self.settingsSection(), 'sortby'))
        orderBy = sortKeys[0] if sortKeys else 'budget'
        self.orderBy(sortKeys[0] if sortKeys else 'budget')
        pub.subscribe(self.on_order_by_changed, 
                      'settings.%s.sortby' % self.settingsSection())
        self.orderUICommand.setChoice(self.__orderBy)
        for eventType in (task.Task.subjectChangedEventType(),
                          task.Task.dueDateTimeChangedEventType(),
                          task.Task.plannedStartDateTimeChangedEventType(),
                          task.Task.completionDateTimeChangedEventType()):
            if eventType.startswith('pubsub'):
                pub.subscribe(self.onAttributeChanged, eventType)
            else:
                self.registerObserver(self.onAttributeChanged_Deprecated, eventType)

    def curselectionIsInstanceOf(self, class_):
        return class_ == task.Task
    
    def createWidget(self):
        itemPopupMenu = self.createTaskPopupMenu()
        self._popupMenus.append(itemPopupMenu)
        return widgets.SquareMap(self, SquareMapRootNode(self.presentation()), 
            self.onSelect, uicommand.Edit(viewer=self), itemPopupMenu)
        
    def createModeToolBarUICommands(self):
        self.orderUICommand = uicommand.SquareTaskViewerOrderChoice(viewer=self,
                                                                    settings=self.settings)  # pylint: disable=W0201
        return super(SquareTaskViewer, self).createModeToolBarUICommands() + \
            (self.orderUICommand,)
        
    def hasModes(self):
        return True
    
    def getModeUICommands(self):
        return [_('Lay out tasks by'), None] + \
            [uicommand.SquareTaskViewerOrderByOption(menuText=menuText,
                                                     value=value, viewer=self, 
                                                     settings=self.settings)
             for (menuText, value) in zip(uicommand.SquareTaskViewerOrderChoice.choiceLabels,
                                          uicommand.SquareTaskViewerOrderChoice.choiceData)]
    
    def on_order_by_changed(self, value):
        self.orderBy(value)
        
    def orderBy(self, choice):
        if choice == self.__orderBy:
            return
        oldChoice = self.__orderBy
        self.__orderBy = choice
        try:
            oldEventType = getattr(task.Task, '%sChangedEventType' % oldChoice)()
        except AttributeError:
            oldEventType = 'task.%s' % oldChoice
        if oldEventType.startswith('pubsub'):
            try:
                pub.unsubscribe(self.onAttributeChanged, oldEventType)
            except pub.TopicNameError:
                pass  # Can happen on first call to orderBy
        else:
            self.removeObserver(self.onAttributeChanged_Deprecated, oldEventType)
        try:
            newEventType = getattr(task.Task, '%sChangedEventType' % choice)()
        except AttributeError:
            newEventType = 'task.%s' % choice
        if newEventType.startswith('pubsub'):
            pub.subscribe(self.onAttributeChanged, newEventType)
        else:
            self.registerObserver(self.onAttributeChanged_Deprecated, newEventType)
        if choice in ('budget', 'timeSpent'):
            self.__transformTaskAttribute = lambda timeSpent: timeSpent.milliseconds() / 1000
            self.__zero = date.TimeDelta()
        else:
            self.__transformTaskAttribute = lambda x: x
            self.__zero = 0
        self.refresh()
        
    def curselection(self):
        # Override curselection, because there is no need to translate indices
        # back to domain objects. Our widget already returns the selected domain
        # object itself.
        return self.widget.curselection()
    
    def nrOfVisibleTasks(self):
        return len([eachTask for eachTask in self.presentation() \
                    if getattr(eachTask, self.__orderBy)(recursive=True) > self.__zero])
        
    # SquareMap adapter methods:
    # pylint: disable=W0621
    
    def overall(self, task):
        return self.__transformTaskAttribute(max(getattr(task, self.__orderBy)(recursive=True),
                                                 self.__zero))
    
    def children_sum(self, children, parent):  # pylint: disable=W0613
        children_sum = sum((max(getattr(child, self.__orderBy)(recursive=True), self.__zero) for child in children \
                            if child in self.presentation()), self.__zero)
        return self.__transformTaskAttribute(max(children_sum, self.__zero))
    
    def empty(self, task):
        overall = self.overall(task)
        if overall:
            children_sum = self.children_sum(self.children(task), task)
            return max(self.__transformTaskAttribute(self.__zero), 
                       (overall - children_sum)) / float(overall)
        return 0
    
    def getItemText(self, task):
        text = super(SquareTaskViewer, self).getItemText(task)
        value = self.render(getattr(task, self.__orderBy)(recursive=False))
        return '%s (%s)' % (text, value) if value else text

    def value(self, task, parent=None):  # pylint: disable=W0613
        return self.overall(task)

    def foreground_color(self, task, depth):  # pylint: disable=W0613
        return task.foregroundColor(recursive=True)
        
    def background_color(self, task, depth):  # pylint: disable=W0613
        red = blue = 255 - (depth * 3) % 100
        green = 255 - (depth * 2) % 100
        color = wx.Colour(red, green, blue)
        return task.backgroundColor(recursive=True) or color
    
    def font(self, task, depth):  # pylint: disable=W0613
        return task.font(recursive=True)

    def icon(self, task, isSelected):
        bitmap = self.iconName(task, isSelected) or 'led_blue_icon'
        return wx.ArtProvider_GetIcon(bitmap, wx.ART_MENU, (16, 16))

    # Helper methods
    
    def render(self, value):
        return self.renderer[self.__orderBy](value)


class HierarchicalCalendarViewer(mixin.AttachmentDropTargetMixin,
                                 mixin.SortableViewerForTasksMixin,
                                 BaseTaskTreeViewer):
    defaultTitle = _('Hierarchical calendar')
    defaultBitmap = 'calendar_icon'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'hierarchicalcalendarviewer')
        super(HierarchicalCalendarViewer, self).__init__(*args, **kwargs)

        # pylint: disable=E1101
        for eventType in (task.Task.subjectChangedEventType(), 
                          task.Task.attachmentsChangedEventType(),
                          task.Task.notesChangedEventType(),
                          task.Task.trackingChangedEventType(),
                          task.Task.percentageCompleteChangedEventType()):
            if eventType.startswith('pubsub'):
                pub.subscribe(self.onAttributeChanged, eventType)
            else:
                self.registerObserver(self.onAttributeChanged_Deprecated, 
                                      eventType)

        # Dates are treated separately because the layout may change (_Invalidate)
        # pylint: disable=E1101
        for eventType in (task.Task.plannedStartDateTimeChangedEventType(),
                          task.Task.dueDateTimeChangedEventType(), 
                          task.Task.completionDateTimeChangedEventType()):
            if eventType.startswith('pubsub'):
                pub.subscribe(self.onLayoutAttributeChanged, eventType)
            else:
                self.registerObserver(self.onLayoutAttributeChanged_Deprecated, 
                                      eventType)

        self.reconfig()

        date.Scheduler().schedule_interval(self.atMidnight, days=1)

    def reconfig(self):
        self.widget.SetCalendarFormat(self.settings.getint(self.settingsSection(), 'calendarformat'))
        self.widget.SetHeaderFormat(self.settings.getint(self.settingsSection(), 'headerformat'))
        self.widget.SetDrawNow(self.settings.getboolean(self.settingsSection(), 'drawnow'))
        self.widget.SetTodayColor(map(int, self.settings.get(self.settingsSection(), 'todaycolor').split(',')))

    def configure(self):
        dialog = HierarchicalCalendarConfigDialog(self.settings, self.settingsSection(), 
                                                  self, 
                                                  title=_('Hierarchical calendar viewer configuration'))
        dialog.CentreOnParent()
        if dialog.ShowModal() == wx.ID_OK:
            self.reconfig()

    def createModeToolBarUICommands(self):
        return super(HierarchicalCalendarViewer, self).createModeToolBarUICommands() + \
            (None, uicommand.HierarchicalCalendarViewerConfigure(viewer=self),
             uicommand.HierarchicalCalendarViewerPreviousPeriod(viewer=self),
             uicommand.HierarchicalCalendarViewerToday(viewer=self),
             uicommand.HierarchicalCalendarViewerNextPeriod(viewer=self))

    def detach(self):
        super(HierarchicalCalendarViewer, self).detach()
        date.Scheduler().unschedule(self.atMidnight)

    def atMidnight(self):
        self.widget.SetCalendarFormat(self.widget.CalendarFormat())

    def onLayoutAttributeChanged(self, newValue, sender):
        self.refresh()

    def onLayoutAttributeChanged_Deprecated(self, event):
        self.refresh()

    def isTreeViewer(self):
        return True

    def onEverySecond(self, event):  # pylint: disable=W0221,W0613
        pass

    def createWidget(self):
        itemPopupMenu = self.createTaskPopupMenu()
        self._popupMenus.append(itemPopupMenu)
        widget = widgets.HierarchicalCalendar(self, self.presentation(),
                                  self.onSelect, self.onEdit, self.onCreate, 
                                  itemPopupMenu,
                                  **self.widgetCreationKeywordArguments())
        return widget

    def onEdit(self, item):
        edit = uicommand.Edit(viewer=self)
        edit(item)

    def onCreate(self, dateTime, show=True):
        plannedStartDateTime = dateTime
        dueDateTime = dateTime.endOfDay() if dateTime == dateTime.startOfDay() else dateTime
        create = uicommand.TaskNew(taskList=self.presentation(), 
                                   settings=self.settings,
                                   taskKeywords=dict(plannedStartDateTime=plannedStartDateTime, 
                                                     dueDateTime=dueDateTime))
        return create(event=None, show=show)

    def isAnyItemExpandable(self):
        return False

    def isAnyItemCollapsable(self):
        return False

    def GetPrintout(self, settings):
        return self.widget.GetPrintout(settings)


class CalendarViewer(mixin.AttachmentDropTargetMixin,
                     mixin.SortableViewerForTasksMixin,
                     BaseTaskTreeViewer):
    defaultTitle = _('Calendar')
    defaultBitmap = 'calendar_icon'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'calendarviewer')
        kwargs['doRefresh'] = False
        super(CalendarViewer, self).__init__(*args, **kwargs)

        start = self.settings.get(self.settingsSection(), 'viewdate')
        if start:
            dt = wx.DateTime.Now()
            dt.ParseDateTime(start)
            self.widget.SetDate(dt)

        self.onWeekStartChanged(self.settings.gettext('view', 'weekstart'))
        self.onWorkingHourChanged()

        self.reconfig()
        self.widget.SetPeriodWidth(self.settings.getint(self.settingsSection(), 
                                                        'periodwidth'))

        for eventType in ('start', 'end'):
            pub.subscribe(self.onWorkingHourChanged, 
                          'settings.view.efforthour%s' % eventType)
        pub.subscribe(self.onWeekStartChanged, 'settings.view.weekstartmonday')

        # pylint: disable=E1101
        for eventType in (task.Task.subjectChangedEventType(),
                          task.Task.plannedStartDateTimeChangedEventType(),
                          task.Task.dueDateTimeChangedEventType(),
                          task.Task.completionDateTimeChangedEventType(),
                          task.Task.attachmentsChangedEventType(),
                          task.Task.notesChangedEventType(),
                          task.Task.trackingChangedEventType(),
                          task.Task.percentageCompleteChangedEventType()):
            if eventType.startswith('pubsub'):
                pub.subscribe(self.onAttributeChanged, eventType)
            else:
                self.registerObserver(self.onAttributeChanged_Deprecated, 
                                      eventType)
        date.Scheduler().schedule_interval(self.atMidnight, days=1)

    def detach(self):
        super(CalendarViewer, self).detach()
        date.Scheduler().unschedule(self.atMidnight)

    def isTreeViewer(self):
        return False

    def onEverySecond(self, event):  # pylint: disable=W0221,W0613
        pass  # Too expensive

    def atMidnight(self):
        if not self.settings.get(self.settingsSection(), 'viewdate'):
            # User has selected the "current" date/time; it may have
            # changed now
            self.SetViewType(wxSCHEDULER_TODAY)

    def onWorkingHourChanged(self, value=None):  # pylint: disable=W0613
        self.widget.SetWorkHours(self.settings.getint('view', 'efforthourstart'),
                                 self.settings.getint('view', 'efforthourend'))
        
    def onWeekStartChanged(self, value):
        assert value in ('monday', 'sunday')
        if value == 'monday':
            self.widget.SetWeekStartMonday()
        else:
            self.widget.SetWeekStartSunday()

    def createWidget(self):
        itemPopupMenu = self.createTaskPopupMenu()
        self._popupMenus.append(itemPopupMenu)
        widget = widgets.Calendar(self, self.presentation(), self.iconName, 
                                  self.onSelect, self.onEdit, self.onCreate, 
                                  self.onChangeConfig, itemPopupMenu,
                                  **self.widgetCreationKeywordArguments())

        if self.settings.getboolean('calendarviewer', 'gradient'):
            # If called directly, we crash with a Cairo assert failing...
            wx.CallAfter(widget.SetDrawer, wxFancyDrawer)

        return widget

    def onChangeConfig(self):
        self.settings.set(self.settingsSection(), 'periodwidth', 
                          str(self.widget.GetPeriodWidth()))

    def onEdit(self, item):
        edit = uicommand.Edit(viewer=self)
        edit(item)

    def onCreate(self, dateTime, show=True):
        plannedStartDateTime = dateTime
        dueDateTime = dateTime.endOfDay() if dateTime == dateTime.startOfDay() else dateTime
        create = uicommand.TaskNew(taskList=self.presentation(), 
                                   settings=self.settings,
                                   taskKeywords=dict(plannedStartDateTime=plannedStartDateTime, 
                                                     dueDateTime=dueDateTime))
        return create(event=None, show=show)

    def createModeToolBarUICommands(self):
        return super(CalendarViewer, self).createModeToolBarUICommands() + \
            (None, uicommand.CalendarViewerConfigure(viewer=self),
             uicommand.CalendarViewerPreviousPeriod(viewer=self),
             uicommand.CalendarViewerToday(viewer=self),
             uicommand.CalendarViewerNextPeriod(viewer=self))

    def SetViewType(self, type_):
        self.widget.SetViewType(type_)
        dt = self.widget.GetDate()
        now = wx.DateTime.Today()
        if (dt.GetYear(), dt.GetMonth(), dt.GetDay()) == (now.GetYear(), now.GetMonth(), now.GetDay()):
            toSave = ''
        else:
            toSave = dt.Format()
        self.settings.set(self.settingsSection(), 'viewdate', toSave)

    # We need to override these because BaseTaskTreeViewer is a tree viewer, but
    # CalendarViewer is not. There is probably a better solution...

    def isAnyItemExpandable(self):
        return False

    def isAnyItemCollapsable(self):
        return False

    def reconfig(self):
        self.widget.Freeze()
        try:
            self.widget.SetPeriodCount(self.settings.getint(self.settingsSection(), 'periodcount'))
            self.widget.SetViewType(self.settings.getint(self.settingsSection(), 'viewtype'))
            self.widget.SetStyle(self.settings.getint(self.settingsSection(), 'vieworientation'))
            self.widget.SetShowNoStartDate(self.settings.getboolean(self.settingsSection(), 'shownostart'))
            self.widget.SetShowNoDueDate(self.settings.getboolean(self.settingsSection(), 'shownodue'))
            self.widget.SetShowUnplanned(self.settings.getboolean(self.settingsSection(), 'showunplanned'))
            self.widget.SetShowNow(self.settings.getboolean(self.settingsSection(), 'shownow'))

            hcolor = self.settings.get(self.settingsSection(), 'highlightcolor')
            if hcolor:
                highlightColor = wx.Colour(*tuple([int(c) for c in hcolor.split(',')]))
                self.widget.SetHighlightColor(highlightColor)
            self.widget.RefreshAllItems(0)
        finally:
            self.widget.Thaw()

    def configure(self):
        dialog = CalendarConfigDialog(self.settings, self.settingsSection(), 
                                      self, 
                                      title=_('Calendar viewer configuration'))
        dialog.CentreOnParent()
        if dialog.ShowModal() == wx.ID_OK:
            self.reconfig()

    def GetPrintout(self, settings):
        return self.widget.GetPrintout(settings)


class TaskViewer(mixin.AttachmentDropTargetMixin,  # pylint: disable=W0223
                 mixin.SortableViewerForTasksMixin, 
                 mixin.NoteColumnMixin, mixin.AttachmentColumnMixin,
                 base.SortableViewerWithColumns, BaseTaskTreeViewer):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'taskviewer')
        super(TaskViewer, self).__init__(*args, **kwargs)
        if self.isVisibleColumnByName('timeLeft'):
            self.minuteRefresher.startClock()
        pub.subscribe(self.onTreeListModeChanged, 
                      'settings.%s.treemode' % self.settingsSection())

    def activate(self):
        if hasattr(wx.GetTopLevelParent(self), 'AddBalloonTip'):
            wx.GetTopLevelParent(self).AddBalloonTip(self.settings, 'manualordering', self.widget,
                title=_('Manual ordering'),
                getRect=lambda: wx.Rect(0, 0, 28, 16),
                message=_('''Show the "Manual ordering" column, then drag and drop items from this column to sort them arbitrarily.'''))

    def isTreeViewer(self):
        # We first ask our presentation what the mode is because 
        # ConfigParser.getboolean is a relatively expensive method. However,
        # when initializing, the presentation might not be created yet. So in
        # that case we get an AttributeError and we use the settings.
        try:
            return self.presentation().treeMode()
        except AttributeError:
            return self.settings.getboolean(self.settingsSection(), 'treemode')

    def showColumn(self, column, show=True, *args, **kwargs):
        if column.name() == 'timeLeft':
            if show:
                self.minuteRefresher.startClock()
            else:
                self.minuteRefresher.stopClock()
        super(TaskViewer, self).showColumn(column, show, *args, **kwargs)

    def curselectionIsInstanceOf(self, class_):
        return class_ == task.Task
    
    def createWidget(self):
        imageList = self.createImageList()  # Has side-effects
        self._columns = self._createColumns()
        itemPopupMenu = self.createTaskPopupMenu()
        columnPopupMenu = self.createColumnPopupMenu()
        self._popupMenus.extend([itemPopupMenu, columnPopupMenu])
        widget = widgets.TreeListCtrl(self, self.columns(), self.onSelect, 
            uicommand.Edit(viewer=self),
            uicommand.TaskDragAndDrop(taskList=self.presentation(), 
                                      viewer=self),
            itemPopupMenu, columnPopupMenu,
            resizeableColumn=1 if self.hasOrderingColumn() else 0,
            validateDrag=self.validateDrag,
            **self.widgetCreationKeywordArguments())
        if self.hasOrderingColumn():
            widget.SetMainColumn(1)
        widget.AssignImageList(imageList)  # pylint: disable=E1101
        widget.Bind(wx.EVT_TREE_BEGIN_LABEL_EDIT, self.onBeginEdit)
        widget.Bind(wx.EVT_TREE_END_LABEL_EDIT, self.onEndEdit)
        return widget

    def onBeginEdit(self, event):
        ''' Make sure only the non-recursive part of the subject can be
            edited inline. '''
        event.Skip()
        if not self.isTreeViewer():
            # Make sure the text control only shows the non-recursive subject
            # by temporarily changing the item text into the non-recursive
            # subject. When the editing ends, we change the item text back into
            # the recursive subject. See onEndEdit.
            treeItem = event.GetItem()
            editedTask = self.widget.GetItemPyData(treeItem)
            self.widget.SetItemText(treeItem, editedTask.subject())
            
    def onEndEdit(self, event):
        ''' Make sure only the non-recursive part of the subject can be
            edited inline. '''
        event.Skip()
        if not self.isTreeViewer():
            # Restore the recursive subject. Here we don't care whether users
            # actually changed the subject. If they did, the subject will 
            # be updated via the regular notification mechanism.
            treeItem = event.GetItem()
            editedTask = self.widget.GetItemPyData(treeItem)
            self.widget.SetItemText(treeItem, 
                                    editedTask.subject(recursive=True))
    
    def _createColumns(self):
        kwargs = dict(resizeCallback=self.onResizeColumn)
        # pylint: disable=E1101,W0142
        columns = [
            widgets.Column('ordering', u'',
                task.Task.orderingChangedEventType(),
                sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                    value='ordering'),
                renderCallback=lambda task: '',
                imageIndicesCallback=self.orderingImageIndices,
                width=self.getColumnWidth('ordering')),
            widgets.Column('subject', _('Subject'), 
                task.Task.subjectChangedEventType(),
                task.Task.completionDateTimeChangedEventType(),
                task.Task.actualStartDateTimeChangedEventType(),
                task.Task.dueDateTimeChangedEventType(),
                task.Task.plannedStartDateTimeChangedEventType(),
                task.Task.trackingChangedEventType(),
                sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                    value='subject'),
                width=self.getColumnWidth('subject'), 
                imageIndicesCallback=self.subjectImageIndices,
                renderCallback=self.renderSubject, 
                editCallback=self.onEditSubject, 
                editControl=inplace_editor.SubjectCtrl, **kwargs)] + \
            [widgets.Column('description', _('Description'), 
                task.Task.descriptionChangedEventType(), 
                sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                    value='description'),
                renderCallback=lambda task: task.description(), 
                width=self.getColumnWidth('description'),  
                editCallback=self.onEditDescription, 
                editControl=inplace_editor.DescriptionCtrl, **kwargs)] + \
            [widgets.Column('attachments', _('Attachments'), 
                task.Task.attachmentsChangedEventType(), 
                width=self.getColumnWidth('attachments'),
                alignment=wx.LIST_FORMAT_LEFT,
                imageIndicesCallback=self.attachmentImageIndices,
                headerImageIndex=self.imageIndex['paperclip_icon'],
                renderCallback=lambda task: '', **kwargs)]
        columns.append(widgets.Column('notes', _('Notes'), 
            task.Task.notesChangedEventType(),
            width=self.getColumnWidth('notes'),
            alignment=wx.LIST_FORMAT_LEFT,
            imageIndicesCallback=self.noteImageIndices,
            headerImageIndex=self.imageIndex['note_icon'],
            renderCallback=lambda task: '', **kwargs))
        columns.extend(
            [widgets.Column('categories', _('Categories'), 
                task.Task.categoryAddedEventType(), 
                task.Task.categoryRemovedEventType(), 
                task.Task.categorySubjectChangedEventType(),
                task.Task.expansionChangedEventType(),
                sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                                                           value='categories'),
                width=self.getColumnWidth('categories'),
                renderCallback=self.renderCategories, **kwargs),
             widgets.Column('prerequisites', _('Prerequisites'),
                task.Task.prerequisitesChangedEventType(), 
                task.Task.expansionChangedEventType(),
                sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                                                           value='prerequisites'),
                renderCallback=self.renderPrerequisites,
                width=self.getColumnWidth('prerequisites'), **kwargs),
             widgets.Column('dependencies', _('Dependents'),
                task.Task.dependenciesChangedEventType(), 
                task.Task.expansionChangedEventType(),
                sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                                                           value='dependencies'),
                renderCallback=self.renderDependencies,
                width=self.getColumnWidth('dependencies'), **kwargs)])

        for name, columnHeader, editCtrl, editCallback, eventTypes in [
            ('plannedStartDateTime', _('Planned start date'), 
             inplace_editor.DateTimeCtrl, self.onEditPlannedStartDateTime, []),
            ('dueDateTime', _('Due date'), DueDateTimeCtrl, 
             self.onEditDueDateTime, [task.Task.expansionChangedEventType()]),
            ('actualStartDateTime', _('Actual start date'), 
             inplace_editor.DateTimeCtrl, self.onEditActualStartDateTime, 
             [task.Task.expansionChangedEventType()]),
            ('completionDateTime', _('Completion date'), 
             inplace_editor.DateTimeCtrl, self.onEditCompletionDateTime, 
             [task.Task.expansionChangedEventType()])]:
            renderCallback = getattr(self, 'render%s' % \
                                     (name[0].capitalize() + name[1:]))
            columns.append(widgets.Column(name, columnHeader,  
                sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                                                           value=name),
                renderCallback=renderCallback, width=self.getColumnWidth(name),
                alignment=wx.LIST_FORMAT_RIGHT, editControl=editCtrl, 
                editCallback=editCallback, settings=self.settings, *eventTypes, 
                **kwargs))

        dependsOnEffortFeature = ['budget', 'timeSpent', 'budgetLeft',
                                  'hourlyFee', 'fixedFee', 'revenue']
            
        for name, columnHeader, editCtrl, editCallback, eventTypes in [
            ('percentageComplete', _('% complete'), 
             inplace_editor.PercentageCtrl, self.onEditPercentageComplete, 
             [task.Task.expansionChangedEventType(), 
              task.Task.percentageCompleteChangedEventType()]),
            ('timeLeft', _('Time left'), None, None, 
             [task.Task.expansionChangedEventType(), 'task.timeLeft']),
            ('recurrence', _('Recurrence'), None, None, 
             [task.Task.expansionChangedEventType(), 
              task.Task.recurrenceChangedEventType()]),
            ('budget', _('Budget'), inplace_editor.BudgetCtrl, 
             self.onEditBudget, [task.Task.expansionChangedEventType(), 
                                 task.Task.budgetChangedEventType()]),
            ('timeSpent', _('Time spent'), None, None, 
             [task.Task.expansionChangedEventType(), 
              task.Task.timeSpentChangedEventType()]),
            ('budgetLeft', _('Budget left'), None, None, 
             [task.Task.expansionChangedEventType(), 
              task.Task.budgetLeftChangedEventType()]),            
            ('priority', _('Priority'), inplace_editor.PriorityCtrl, 
             self.onEditPriority, [task.Task.expansionChangedEventType(), 
                                   task.Task.priorityChangedEventType()]),
            ('hourlyFee', _('Hourly fee'), inplace_editor.AmountCtrl, 
             self.onEditHourlyFee, [task.Task.hourlyFeeChangedEventType()]),
            ('fixedFee', _('Fixed fee'), inplace_editor.AmountCtrl, 
             self.onEditFixedFee, [task.Task.expansionChangedEventType(), 
                                   task.Task.fixedFeeChangedEventType()]),
            ('revenue', _('Revenue'), None, None, 
             [task.Task.expansionChangedEventType(), 
              task.Task.revenueChangedEventType()])]:
            if (name in dependsOnEffortFeature) or name not in dependsOnEffortFeature:
                renderCallback = getattr(self, 'render%s' % \
                                         (name[0].capitalize() + name[1:]))
                columns.append(widgets.Column(name, columnHeader,  
                    sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                                                               value=name),
                    renderCallback=renderCallback, 
                    width=self.getColumnWidth(name),
                    alignment=wx.LIST_FORMAT_RIGHT, editControl=editCtrl, 
                    editCallback=editCallback, *eventTypes, **kwargs))
                
        columns.append(widgets.Column('reminder', _('Reminder'), 
            sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                                                       value='reminder'),
            renderCallback=self.renderReminder, 
            width=self.getColumnWidth('reminder'),
            alignment=wx.LIST_FORMAT_RIGHT, 
            editControl=inplace_editor.DateTimeCtrl,
            editCallback=self.onEditReminderDateTime, settings=self.settings,
            *[task.Task.expansionChangedEventType(), 
              task.Task.reminderChangedEventType()], **kwargs))
        columns.append(widgets.Column('creationDateTime', _('Creation date'),
            width=self.getColumnWidth('creationDateTime'),
            renderCallback=self.renderCreationDateTime,
            sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                                                       value='creationDateTime'),
            **kwargs))
        columns.append(widgets.Column('modificationDateTime', 
                                      _('Modification date'),
            width=self.getColumnWidth('modificationDateTime'),
            renderCallback=self.renderModificationDateTime,
            sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                                                       value='modificationDateTime'),
            *task.Task.modificationEventTypes(), **kwargs))
        return columns
    
    def createColumnUICommands(self):
        commands = [
            uicommand.ToggleAutoColumnResizing(viewer=self,
                                               settings=self.settings),
            None,
            (_('&Dates'),
             uicommand.ViewColumns(menuText=_('&All date columns'),
                helpText=_('Show/hide all date-related columns'),
                setting=['plannedStartDateTime', 'dueDateTime', 'timeLeft', 
                         'actualStartDateTime', 'completionDateTime', 
                         'recurrence'],
                viewer=self),
             None,
             uicommand.ViewColumn(menuText=_('&Planned start date'),
                 helpText=_('Show/hide planned start date column'),
                 setting='plannedStartDateTime', viewer=self),
             uicommand.ViewColumn(menuText=_('&Due date'),
                 helpText=_('Show/hide due date column'),
                 setting='dueDateTime', viewer=self),
             uicommand.ViewColumn(menuText=_('&Actual start date'),
                helpText=_('Show/hide actual start date column'),
                setting='actualStartDateTime', viewer=self),
             uicommand.ViewColumn(menuText=_('&Completion date'),
                 helpText=_('Show/hide completion date column'),
                 setting='completionDateTime', viewer=self),
             uicommand.ViewColumn(menuText=_('&Time left'),
                 helpText=_('Show/hide time left column'),
                 setting='timeLeft', viewer=self),
             uicommand.ViewColumn(menuText=_('&Recurrence'),
                 helpText=_('Show/hide recurrence column'),
                 setting='recurrence', viewer=self))]
        commands.extend([
            (_('&Budget'),
             uicommand.ViewColumns(menuText=_('&All budget columns'),
                 helpText=_('Show/hide all budget-related columns'),
                 setting=['budget', 'timeSpent', 'budgetLeft'],
                 viewer=self),
             None,
             uicommand.ViewColumn(menuText=_('&Budget'),
                 helpText=_('Show/hide budget column'),
                 setting='budget', viewer=self),
             uicommand.ViewColumn(menuText=_('&Time spent'),
                 helpText=_('Show/hide time spent column'),
                 setting='timeSpent', viewer=self),
             uicommand.ViewColumn(menuText=_('&Budget left'),
                 helpText=_('Show/hide budget left column'),
                 setting='budgetLeft', viewer=self),
            ),
            (_('&Financial'),
             uicommand.ViewColumns(menuText=_('&All financial columns'),
                 helpText=_('Show/hide all finance-related columns'),
                 setting=['hourlyFee', 'fixedFee', 'revenue'],
                 viewer=self),
             None,
             uicommand.ViewColumn(menuText=_('&Hourly fee'),
                 helpText=_('Show/hide hourly fee column'),
                 setting='hourlyFee', viewer=self),
             uicommand.ViewColumn(menuText=_('&Fixed fee'),
                 helpText=_('Show/hide fixed fee column'),
                 setting='fixedFee', viewer=self),
             uicommand.ViewColumn(menuText=_('&Revenue'),
                 helpText=_('Show/hide revenue column'),
                 setting='revenue', viewer=self))])
        commands.extend([
            uicommand.ViewColumn(menuText=_('&Manual ordering'),
                helpText=_('Show/hide the manual ordering column'),
                setting='ordering', viewer=self),
            uicommand.ViewColumn(menuText=_('&Description'),
                helpText=_('Show/hide description column'),
                setting='description', viewer=self),
            uicommand.ViewColumn(menuText=_('&Prerequisites'),
                 helpText=_('Show/hide prerequisites column'),
                 setting='prerequisites', viewer=self),
            uicommand.ViewColumn(menuText=_('&Dependents'),
                 helpText=_('Show/hide dependents column'),
                 setting='dependencies', viewer=self),
             uicommand.ViewColumn(menuText=_('&Percentage complete'),
                 helpText=_('Show/hide percentage complete column'),
                 setting='percentageComplete', viewer=self),
            uicommand.ViewColumn(menuText=_('&Attachments'),
                helpText=_('Show/hide attachment column'),
                setting='attachments', viewer=self)])
        commands.append(
            uicommand.ViewColumn(menuText=_('&Notes'),
                helpText=_('Show/hide notes column'),
                setting='notes', viewer=self))
        commands.extend([
            uicommand.ViewColumn(menuText=_('&Categories'),
                helpText=_('Show/hide categories column'),
                setting='categories', viewer=self),
            uicommand.ViewColumn(menuText=_('&Priority'),
                helpText=_('Show/hide priority column'),
                setting='priority', viewer=self),
            uicommand.ViewColumn(menuText=_('&Reminder'),
                helpText=_('Show/hide reminder column'),
                setting='reminder', viewer=self),
            uicommand.ViewColumn(menuText=_('&Creation date'),
                helpText=_('Show/hide creation date column'),
                setting='creationDateTime', viewer=self),
            uicommand.ViewColumn(menuText=_('&Modification date'),
                helpText=_('Show/hide last modification date column'),
                setting='modificationDateTime', viewer=self)])
        return commands

    def createModeToolBarUICommands(self):
        treeOrListUICommand = uicommand.TaskViewerTreeOrListChoice(viewer=self,
                                                                   settings=self.settings)  # pylint: disable=W0201 
        return super(TaskViewer, self).createModeToolBarUICommands() + \
            (treeOrListUICommand,)
            
    def hasModes(self):
        return True
    
    def getModeUICommands(self):
        return [_('Show tasks as'), None] + \
            [uicommand.TaskViewerTreeOrListOption(menuText=menuText, value=value,
                                                  viewer=self, settings=self.settings)
             for (menuText, value) in zip(uicommand.TaskViewerTreeOrListChoice.choiceLabels,
                                          uicommand.TaskViewerTreeOrListChoice.choiceData)]

    def createColumnPopupMenu(self):
        return menu.ColumnPopupMenu(self)
                    
    def setSortByTaskStatusFirst(self, *args, **kwargs):  # pylint: disable=W0221
        super(TaskViewer, self).setSortByTaskStatusFirst(*args, **kwargs)
        self.showSortOrder()
        
    def getSortOrderImage(self):
        sortOrderImage = super(TaskViewer, self).getSortOrderImage()
        if self.isSortByTaskStatusFirst():  # pylint: disable=E1101
            sortOrderImage = sortOrderImage.rstrip('icon') + 'with_status_icon'
        return sortOrderImage

    def setSearchFilter(self, searchString, *args, **kwargs):  # pylint: disable=W0221
        super(TaskViewer, self).setSearchFilter(searchString, *args, **kwargs)
        if searchString:
            self.expandAll()  # pylint: disable=E1101      

    def onTreeListModeChanged(self, value):
        self.presentation().setTreeMode(value)
        
    # pylint: disable=W0621
    
    def renderSubject(self, task):
        return task.subject(recursive=not self.isTreeViewer())
    
    def renderPlannedStartDateTime(self, task, humanReadable=True):
        return self.renderedValue(task, task.plannedStartDateTime,
                                  lambda x: render.dateTime(x, humanReadable=humanReadable))
    
    def renderDueDateTime(self, task, humanReadable=True):
        return self.renderedValue(task, task.dueDateTime,
                                  lambda x: render.dateTime(x, humanReadable=humanReadable))

    def renderActualStartDateTime(self, task, humanReadable=True):
        return self.renderedValue(task, task.actualStartDateTime, 
                                  lambda x: render.dateTime(x, humanReadable=humanReadable))

    def renderCompletionDateTime(self, task, humanReadable=True):
        return self.renderedValue(task, task.completionDateTime, 
                                  lambda x: render.dateTime(x, humanReadable=humanReadable))

    def renderRecurrence(self, task):
        return self.renderedValue(task, task.recurrence, render.recurrence)
    
    def renderPrerequisites(self, task):
        return self.renderSubjectsOfRelatedItems(task, task.prerequisites)
    
    def renderDependencies(self, task):
        return self.renderSubjectsOfRelatedItems(task, task.dependencies)
    
    def renderTimeLeft(self, task):
        return self.renderedValue(task, task.timeLeft, render.timeLeft, 
                                  task.completed())
        
    def renderTimeSpent(self, task):
        return self.renderedValue(task, task.timeSpent, self._renderTimeSpent)

    def renderBudget(self, task):
        return self.renderedValue(task, task.budget, render.budget)

    def renderBudgetLeft(self, task):
        return self.renderedValue(task, task.budgetLeft, render.budget)

    def renderRevenue(self, task):
        return self.renderedValue(task, task.revenue, render.monetaryAmount)
    
    def renderHourlyFee(self, task):
        # hourlyFee has no recursive value
        return render.monetaryAmount(task.hourlyFee())  
    
    def renderFixedFee(self, task):
        return self.renderedValue(task, task.fixedFee, render.monetaryAmount)

    def renderPercentageComplete(self, task):
        return self.renderedValue(task, task.percentageComplete, 
                                  render.percentage)

    def renderPriority(self, task):
        return self.renderedValue(task, task.priority, render.priority) + ' '
    
    def renderReminder(self, task, humanReadable=True):
        return self.renderedValue(task, task.reminder,
                                  lambda x: render.dateTime(x, humanReadable=humanReadable))
    
    def renderedValue(self, item, getValue, renderValue, *extraRenderArgs):
        value = getValue(recursive=False)
        template = '%s'
        if self.isItemCollapsed(item):
            recursiveValue = getValue(recursive=True)
            if value != recursiveValue:
                value = recursiveValue
                template = '(%s)'
        return template % renderValue(value, *extraRenderArgs)
    
    def onEditPlannedStartDateTime(self, item, newValue):
        keep_delta = self.settings.get('view', 'datestied') == 'startdue'
        command.EditPlannedStartDateTimeCommand(items=[item], 
                                                newValue=newValue, 
                                                keep_delta=keep_delta).do()
        
    def onEditDueDateTime(self, item, newValue):       
        keep_delta = self.settings.get('view', 'datestied') == 'duestart'
        command.EditDueDateTimeCommand(items=[item], newValue=newValue,
                                       keep_delta=keep_delta).do()

    def onEditActualStartDateTime(self, item, newValue):
        command.EditActualStartDateTimeCommand(items=[item], 
                                               newValue=newValue).do()
        
    def onEditCompletionDateTime(self, item, newValue):
        command.EditCompletionDateTimeCommand(items=[item], 
                                              newValue=newValue).do()
        
    def onEditPercentageComplete(self, item, newValue):
        command.EditPercentageCompleteCommand(items=[item], 
                                              newValue=newValue).do()  # pylint: disable=E1101
        
    def onEditBudget(self, item, newValue):
        command.EditBudgetCommand(items=[item], newValue=newValue).do()
        
    def onEditPriority(self, item, newValue):
        command.EditPriorityCommand(items=[item], newValue=newValue).do()
        
    def onEditReminderDateTime(self, item, newValue):
        command.EditReminderDateTimeCommand(items=[item], 
                                            newValue=newValue).do()
        
    def onEditHourlyFee(self, item, newValue):
        command.EditHourlyFeeCommand(items=[item], newValue=newValue).do()
        
    def onEditFixedFee(self, item, newValue):
        command.EditFixedFeeCommand(items=[item], newValue=newValue).do()

    def onEverySecond(self, event):
        # Only update when a column is visible that changes every second 
        if any([self.isVisibleColumnByName(column) for column in 'timeSpent', 
               'budgetLeft', 'revenue']):
            super(TaskViewer, self).onEverySecond(event)

    def getRootItems(self):
        ''' If the viewer is in tree mode, return the real root items. If the
            viewer is in list mode, return all items. '''
        return super(TaskViewer, self).getRootItems() if \
            self.isTreeViewer() else self.presentation()
    
    def getItemParent(self, item):
        return super(TaskViewer, self).getItemParent(item) if \
            self.isTreeViewer() else None

    def children(self, item=None):
        return super(TaskViewer, self).children(item) if \
            (self.isTreeViewer() or item is None) else []


class CheckableTaskViewer(TaskViewer):  # pylint: disable=W0223
    def createWidget(self):
        imageList = self.createImageList()  # Has side-effects
        self._columns = self._createColumns()
        itemPopupMenu = self.createTaskPopupMenu()
        columnPopupMenu = self.createColumnPopupMenu()
        self._popupMenus.extend([itemPopupMenu, columnPopupMenu])
        widget = widgets.CheckTreeCtrl(self, self.columns(), self.onSelect,
            self.onCheck, uicommand.Edit(viewer=self),
            uicommand.TaskDragAndDrop(taskList=self.presentation(), 
                                      viewer=self),
            itemPopupMenu, columnPopupMenu,
            **self.widgetCreationKeywordArguments())
        widget.AssignImageList(imageList)  # pylint: disable=E1101
        return widget    
    
    def onCheck(self, event, final):
        pass
    
    def getIsItemChecked(self, task):  # pylint: disable=W0613,W0621
        return False
    
    def getItemParentHasExclusiveChildren(self, task):  # pylint: disable=W0613,W0621
        return False
    
    
class TaskStatsViewer(BaseTaskViewer):  # pylint: disable=W0223
    defaultTitle = _('Task statistics')
    defaultBitmap = 'charts_icon'

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'taskstatsviewer')
        super(TaskStatsViewer, self).__init__(*args, **kwargs)
        pub.subscribe(self.onPieChartAngleChanged, 
                      'settings.%s.piechartangle' % self.settingsSection())
        
    def createWidget(self):
        widget = wx.lib.agw.piectrl.PieCtrl(self)
        widget.SetShowEdges(False)
        widget.SetHeight(20)
        self.initLegend(widget)
        for dummy in task.Task.possibleStatuses():
            widget._series.append(wx.lib.agw.piectrl.PiePart(1))  # pylint: disable=W0212
        return widget

    def createClipboardToolBarUICommands(self):
        return ()
    
    def createEditToolBarUICommands(self):
        return ()
    
    def createCreationToolBarUICommands(self):
        return (uicommand.TaskNew(taskList=self.presentation(),
                                  settings=self.settings),
                uicommand.TaskNewFromTemplateButton(taskList=self.presentation(),
                                                    settings=self.settings,
                                                    bitmap='newtmpl'))
        
    def createActionToolBarUICommands(self):
        return tuple([uicommand.ViewerHideTasks(taskStatus=status,
                                                settings=self.settings,
                                                viewer=self) \
                            for status in task.Task.possibleStatuses()]) + \
               (uicommand.ViewerPieChartAngle(viewer=self, 
                                              settings=self.settings),)

    def initLegend(self, widget):
        legend = widget.GetLegend()
        legend.SetTransparent(False)
        legend.SetBackColour(wx.WHITE)
        legend.SetLabelFont(wx.SystemSettings.GetFont(wx.SYS_SYSTEM_FONT))
        legend.Show()
    
    def refresh(self):
        self.widget.SetAngle(self.settings.getint(self.settingsSection(), 
                             'piechartangle') / 180. * math.pi)
        self.refreshParts()
        self.widget.Refresh()
        
    def refreshParts(self):
        series = self.widget._series  # pylint: disable=W0212
        tasks = self.presentation()
        total = len(tasks)
        counts = tasks.nrOfTasksPerStatus()
        for part, status in zip(series, task.Task.possibleStatuses()):
            nrTasks = counts[status]
            percentage = round(100. * nrTasks / total) if total else 0
            part.SetLabel(status.countLabel % (nrTasks, percentage))
            part.SetValue(nrTasks)
            part.SetColour(self.getFgColor(status))
        # PietCtrl can't handle empty pie charts:    
        if total == 0:
            series[0].SetValue(1)
       
    def getFgColor(self, status):
        color = wx.Colour(*eval(self.settings.get('fgcolor', 
                                                  '%stasks' % status)))
        if status == task.status.active and color == wx.BLACK:
            color = wx.BLUE
        return color
        
    def refreshItems(self, *args, **kwargs):  # pylint: disable=W0613
        self.refresh()
    
    def select(self, *args):
        pass
    
    def updateSelection(self, *args, **kwargs):
        pass

    def isTreeViewer(self):
        return False

    def onPieChartAngleChanged(self, value):  # pylint: disable=W0613
        self.refresh()




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns, operating_system
from taskcoachlib.i18n import _
import wx
import icons


class ArtProvider(wx.ArtProvider):
    def CreateBitmap(self, artId, artClient, size):
        if '+' in artId:
            w, h = size
            main, overlay = artId.split('+')

            overlayImage = self._CreateBitmap(overlay, artClient, size).ConvertToImage()
            overlayImage.Rescale(int(w / 2), int(h / 2), wx.IMAGE_QUALITY_HIGH)
            overlayAlpha = overlayImage.GetAlphaData()
            overlayBitmap = overlayImage.ConvertToBitmap()

            mainImage = self._CreateBitmap(main, artClient, size).ConvertToImage()
            mainAlpha = mainImage.GetAlphaData()
            mainImage.SetAlphaData(chr(255) * len(mainAlpha))
            mainBitmap = mainImage.ConvertToBitmap()

            dstDC = wx.MemoryDC()
            dstDC.SelectObject(mainBitmap)
            try:
                dstDC.DrawBitmap(overlayBitmap, w - int(w / 2), h - int(h / 2), True)
            finally:
                dstDC.SelectObject(wx.NullBitmap)
            mainImage = mainBitmap.ConvertToImage()

            # Just drawing works fine on OS X but clips to the destination bitmap on
            # other platforms. There doesn't seem to be anything better than this.
            resultAlpha = list()
            for y in xrange(h):
                for x in xrange(w):
                    alpha = mainAlpha[y * w + x]
                    if x >= w / 2 and y >= h / 2:
                        alpha = max(alpha, overlayAlpha[(y - h / 2) * w / 2 + x - w / 2])
                    resultAlpha.append(alpha)
            mainImage.SetAlphaData(''.join(resultAlpha))

            return mainImage.ConvertToBitmap()
        else:
            return self._CreateBitmap(artId, artClient, size)

    def _CreateBitmap(self, artId, artClient, size):
        if not artId:
            return wx.EmptyBitmap(*size)
        catalogKey = '%s%dx%d' % (artId, size[0], size[1])
        if catalogKey in icons.catalog.keys():
            bitmap = icons.catalog[catalogKey].getBitmap()
            if artClient == wx.ART_FRAME_ICON:
                bitmap = self.convertAlphaToMask(bitmap)
            return bitmap
        else:
            return wx.NullBitmap

    @staticmethod
    def convertAlphaToMask(bitmap):
        image = wx.ImageFromBitmap(bitmap)
        image.ConvertAlphaToMask()
        return wx.BitmapFromImage(image)    


class IconProvider(object):
    __metaclass__ = patterns.Singleton

    def __init__(self):
        self.__iconCache = dict()
        if operating_system.isMac():
            self.__iconSizeOnCurrentPlatform = 128
        elif operating_system.isGTK():
            self.__iconSizeOnCurrentPlatform = 48
        else:
            self.__iconSizeOnCurrentPlatform = 16
        
    def getIcon(self, iconTitle): 
        ''' Return the icon. Use a cache to prevent leakage of GDI object 
            count. '''
        try:
            return self.__iconCache[iconTitle]
        except KeyError:
            icon = self.getIconFromArtProvider(iconTitle)
            self.__iconCache[iconTitle] = icon
            return icon
        
    def iconBundle(self, iconTitle):
        ''' Create an icon bundle with icons of different sizes. '''
        bundle = wx.IconBundle()
        for size in (16, 22, 32, 48, 64, 128):
            bundle.AddIcon(self.getIconFromArtProvider(iconTitle, size))
        return bundle
    
    def getIconFromArtProvider(self, iconTitle, iconSize=None):
        size = iconSize or self.__iconSizeOnCurrentPlatform
        # I just spent two hours trying to get rid of garbage in the icon
        # background on KDE. I give up.
        if operating_system.isGTK():
            return wx.ArtProvider_GetIcon(iconTitle, wx.ART_FRAME_ICON, (size, size))

        # wx.ArtProvider_GetIcon doesn't convert alpha to mask, so we do it
        # ourselves:
        bitmap = wx.ArtProvider_GetBitmap(iconTitle, wx.ART_FRAME_ICON, 
                                          (size, size))    
        bitmap = ArtProvider.convertAlphaToMask(bitmap)
        return wx.IconFromBitmap(bitmap)


def iconBundle(iconTitle):
    return IconProvider().iconBundle(iconTitle)


def getIcon(iconTitle):
    return IconProvider().getIcon(iconTitle)


def init():
    if operating_system.isWindows() and wx.DisplayDepth() >= 32:
        wx.SystemOptions_SetOption("msw.remap", "0")  # pragma: no cover
    try:
        wx.ArtProvider_PushProvider(ArtProvider())  # pylint: disable=E1101
    except AttributeError:
        wx.ArtProvider.Push(ArtProvider())


chooseableItemImages = dict( \
    arrow_down_icon=_('Arrow down'),
    arrow_down_with_status_icon=_('Arrow down with status'),
    arrows_looped_blue_icon=_('Blue arrows looped'),
    arrows_looped_green_icon=_('Green arrows looped'),
    arrow_up_icon=_('Arrow up'),
    arrow_up_with_status_icon=_('Arrow up with status'),
    bomb_icon=_('Bomb'),
    book_icon=_('Book'),
    books_icon=_('Books'),
    box_icon=_('Box'),
    bug_icon=_('Ladybug'),
    cake_icon=_('Cake'),
    calculator_icon=_('Calculator'),
    calendar_icon=_('Calendar'),
    cat_icon=_('Cat'),
    cd_icon=_('Compact disc (CD)'),
    charts_icon=_('Charts'),
    chat_icon=_('Chatting'),
    checkmark_green_icon=_('Check mark'),
    checkmark_green_icon_multiple=_('Check marks'),
    clock_icon=_('Clock'),
    clock_alarm_icon=_('Alarm clock'),
    clock_stopwatch_icon=_('Stopwatch'),
    cogwheel_icon=_('Cogwheel'),
    cogwheels_icon=_('Cogwheels'),
    computer_desktop_icon=_('Desktop computer'),
    computer_laptop_icon=_('Laptop computer'),
    computer_handheld_icon=_('Handheld computer'),
    cross_red_icon=_('Red cross'),
    die_icon=_('Die'),
    document_icon=_('Document'),
    earth_blue_icon=_('Blue earth'),
    earth_green_icon=_('Green earth'),
    envelope_icon=_('Envelope'),
    envelopes_icon=_('Envelopes'),
    folder_blue_icon=_('Blue folder'),
    folder_blue_light_icon=_('Light blue folder'),
    folder_green_icon=_('Green folder'),
    folder_grey_icon=_('Grey folder'),
    folder_orange_icon=_('Orange folder'),
    folder_purple_icon=_('Purple folder'),
    folder_red_icon=_('Red folder'),
    folder_yellow_icon=_('Yellow folder'),
    folder_blue_arrow_icon=_('Blue folder with arrow'),
    heart_icon=_('Heart'),
    hearts_icon=_('Hearts'),
    house_green_icon=_('Green house'),
    house_red_icon=_('Red house'),
    key_icon=_('Key'),
    keys_icon=_('Keys'),
    lamp_icon=_('Lamp'),
    led_blue_questionmark_icon=_('Question mark'),
    led_blue_information_icon=_('Information'),
    led_blue_icon=_('Blue led'),
    led_blue_light_icon=_('Light blue led'),
    led_grey_icon=_('Grey led'),
    led_green_icon=_('Green led'),
    led_green_light_icon=_('Light green led'),
    led_orange_icon=_('Orange led'),
    led_purple_icon=_('Purple led'),
    led_red_icon=_('Red led'),
    led_yellow_icon=_('Yellow led'),
    life_ring_icon=_('Life ring'),
    lock_locked_icon=_('Locked lock'),
    lock_unlocked_icon=_('Unlocked lock'),
    magnifier_glass_icon=_('Magnifier glass'),
    music_piano_icon=_('Piano'),
    music_note_icon=_('Music note'),
    note_icon=_('Note'),
    palette_icon=_('Palette'),
    paperclip_icon=_('Paperclip'),
    pencil_icon=_('Pencil'),
    person_icon=_('Person'),
    persons_icon=_('People'),
    person_id_icon=_('Identification'),
    person_talking_icon=_('Person talking'),
    sign_warning_icon=_('Warning sign'),
    symbol_minus_icon=_('Minus'),
    symbol_plus_icon=_('Plus'),
    star_red_icon=_('Red star'),
    star_yellow_icon=_('Yellow star'),
    trafficlight_icon=_('Traffic light'),
    trashcan_icon=_('Trashcan'),
    weather_lightning_icon=_('Lightning'),
    weather_umbrella_icon=_('Umbrella'),
    weather_sunny_icon=_('Partly sunny'),
    wrench_icon=_('Wrench'))

itemImages = chooseableItemImages.keys() + ['folder_blue_open_icon',
    'folder_green_open_icon', 'folder_grey_open_icon',
    'folder_orange_open_icon', 'folder_red_open_icon',
    'folder_purple_open_icon', 'folder_yellow_open_icon',
    'folder_blue_light_open_icon']

chooseableItemImages[''] = _('No icon')



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import gui, config, persistence
from taskcoachlib.domain import task

       
class TreeViewerTest(test.wxTestCase):
    def setUp(self):
        super(TreeViewerTest, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.viewer = gui.viewer.TaskViewer(self.frame, self.taskFile,
            self.settings)
        self.expansionContext = self.viewer.settingsSection()
        self.parent = task.Task('parent')
        self.child = task.Task('child')
        self.parent.addChild(self.child)
        self.child.setParent(self.parent)
        self.taskFile.tasks().extend([self.parent, self.child])
        self.viewer.refresh()
        self.widget = self.viewer.widget

    def tearDown(self):
        super(TreeViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def firstItem(self):
        root = self.widget.GetRootItem()
        return self.widget.GetFirstChild(root)[0]
                
    def testWidgetDoesNotDisplayChildItemBeforeItsParentIsExpanded(self):
        self.assertEqual(1, self.viewer.widget.GetItemCount())
        
    def testExpand(self):
        self.widget.Expand(self.firstItem())
        self.failUnless(self.parent.isExpanded(context=self.expansionContext))
        
    def testCollapse(self):
        firstVisibleItem = self.firstItem()
        self.widget.Expand(firstVisibleItem)
        self.widget.Collapse(firstVisibleItem)
        self.failIf(self.parent.isExpanded(context=self.expansionContext))
        
    def testExpandall(self):
        self.viewer.expandAll()
        self.failUnless(self.parent.isExpanded(context=self.expansionContext))



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import gui, config, persistence


class PreferencesTest(test.wxTestCase):
    def setUp(self):
        super(PreferencesTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.preferences = gui.Preferences(parent=self.frame, title='Test',
            settings=self.settings)
        self.originalColor = self.settings.get('fgcolor', 'activetasks')
        self.newColor = (1, 2, 29)
        
    # pylint: disable=W0212
    
    def testCancel(self):
        self.preferences[5]._colorSettings[4][2].SetColour(self.newColor)
        self.preferences.cancel()
        self.assertEqual(self.originalColor, self.settings.get('fgcolor', 'activetasks'))
        
    def testOk(self):
        self.preferences[5]._colorSettings[4][2].SetColour(self.newColor)
        self.preferences.ok()
        self.assertEqual(self.newColor, 
            eval(self.settings.get('fgcolor', 'activetasks'))[:3])
        

class SyncMLPreferencesTest(test.TestCase):
    def testCreate(self):
        taskFile = persistence.TaskFile()
        try:
            gui.dialog.syncpreferences.SyncMLPreferences(parent=None,
                  iocontroller=gui.IOController(taskFile, None, None), 
                  title='Edit SyncML preferences')
        finally:
            taskFile.close()
            taskFile.stop()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.notify import NotificationFrameBase, NotificationCenter
from taskcoachlib.i18n import _
import wx


class IPhoneSyncFrame(NotificationFrameBase):
    def __init__(self, settings, *args, **kwargs):
        self.settings = settings

        super(IPhoneSyncFrame, self).__init__(*args, **kwargs)

    def AddInnerContent(self, sizer, panel):
        self.text = wx.StaticText(panel, wx.ID_ANY, _('Synchronizing...'))
        sizer.Add(self.text,
                  0, wx.ALL, 3)

        self.gauge = wx.Gauge(panel, wx.ID_ANY)
        self.gauge.SetRange(100)
        sizer.Add(self.gauge, 0, wx.EXPAND|wx.ALL, 3)

        if self.settings.getboolean('iphone', 'showlog'):
            self.log = wx.TextCtrl(panel, wx.ID_ANY, u'', style=wx.TE_MULTILINE|wx.TE_READONLY)
            sizer.Add(self.log, 1, wx.EXPAND|wx.ALL, 3)

            self.btn = wx.Button(panel, wx.ID_ANY, _('OK'))
            sizer.Add(self.btn, 0, wx.ALIGN_CENTRE|wx.ALL, 3)
            self.btn.Enable(False)
            wx.EVT_BUTTON(self.btn, wx.ID_ANY, self.OnOK)

    def CloseButton(self, panel):
        return None

    def SetDeviceName(self, name):
        self.text.SetLabel(_('Synchronizing with %s...') % name)

    def SetProgress(self, value, total):
        self.gauge.SetValue(int(100 * value / total))

    def AddLogLine(self, line):
        if self.settings.getboolean('iphone', 'showlog'):
            self.log.AppendText(line + u'\n')

    def Started(self):
        NotificationCenter().NotifyFrame(self)

    def Finished(self):
        if self.settings.getboolean('iphone', 'showlog'):
            self.btn.Enable(True)
        else:
            self.DoClose()

    def OnOK(self, event):
        self.DoClose()



# -*- coding: utf-8 -*-

"""
Test for simple escape/unescape routine
"""


from .py30compat import unittest

from keyring.util import escape


class EscapeTestCase(unittest.TestCase):

    def check_escape_unescape(self, initial):
        escaped = escape.escape(initial)
        for c in escaped:
            self.assertIn(c, escape.LEGAL_CHARS + '_')
        unescaped = escape.unescape(escaped)
        self.assertEqual(initial, unescaped)

    def test_escape_unescape(self):
        self.check_escape_unescape("aaaa")
        self.check_escape_unescape("aaaa bbbb cccc")
        self.check_escape_unescape(escape.u("Zażółć gęślą jaźń"))
        self.check_escape_unescape("(((P{{{{'''---; ;; '\"|%^")

    def test_low_byte(self):
        """
        Ensure that encoding low bytes (ordinal less than hex F) encode as
        as three bytes to avoid ambiguity. For example '\n' (hex A) should
        encode as '_0A' and not '_A', the latter of which
        isn't matched by the inverse operation.
        """
        self.check_escape_unescape('\n')
        self.check_escape_unescape('\x000')

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(EscapeTestCase))
    return suite

if __name__ == "__main__":
    unittest.main(defaultTest="test_suite")
